"""
Implements the libc API.
"""

from __future__ import annotations

import os
import re
from pathlib import Path

from elftools.elf.relocation import Relocation

import pwndbg.aglib.elf
import pwndbg.aglib.proc
import pwndbg.aglib.vmmap
import pwndbg.lib.cache
import pwndbg.lib.path

from . import glibc
from . import musl
from . import unknown
from . import util
from .dispatch import LibcProvider
from .dispatch import LibcType
from .dispatch import LibcURLs

# Order is important.
_libc_implementations: tuple[LibcProvider, ...] = (glibc, musl, unknown)


class LibcNotFound(Exception):
    pass


def __check_candidates(
    libc_candidates: list[str], ld_candidates: list[str]
) -> tuple[str | None, str | None, LibcProvider | None]:
    """
    Queries the libc implementations on if any of them claim any libc and ld mappings.

    Returns:
        A tuple (claimed libc mapping, claimed ld mapping, claiming implementation). If noone claimed anything,
        "claiming implementation" will be None. It is possible that exactly one of "claimed libc mapping"
        and "claimed ld mapping" is None.
    """

    def verify_libc_path(path: str) -> tuple[bool, LibcProvider]:
        for impl in _libc_implementations:
            if impl.verify_libc_candidate(path):
                # Someone claims that this makes sense!
                return True, impl
        return False, unknown

    def verify_ld_path(path: str) -> tuple[bool, LibcProvider]:
        for impl in _libc_implementations:
            if impl.verify_ld_candidate(path):
                # Someone claims that this makes sense!
                return True, impl
        return False, unknown

    verified_libc_path: str | None = None
    verified_ld_path: str | None = None
    verified_libc_impl: LibcProvider | None = None

    # See if any libc implementation claims one of the candidate libc mappings.
    for cand in libc_candidates:
        ok, approver = verify_libc_path(cand)
        if ok:
            verified_libc_path = cand
            verified_libc_impl = approver
            break

    # See if any libc implementation claims one of the candidate ld mappings.
    for cand in ld_candidates:
        ok, approver = verify_ld_path(cand)
        if ok:
            # Is there a conflict with the libc verifier?
            if verified_libc_impl is not None and verified_libc_impl.type() != approver.type():
                assert verified_libc_path is not None
                raise LibcNotFound(
                    f"Conflict: {verified_libc_path} is a {verified_libc_impl.type().value} mapping"
                    f" while {cand} is a {approver.type()} mapping."
                )

            verified_ld_path = cand
            verified_libc_impl = approver
            break

    return verified_libc_path, verified_ld_path, verified_libc_impl


libc_regex = re.compile(r"^libc6?[-_\.]")
ld_regex = re.compile(r"ld.*\.so(?:\.[0-9]+)?")

# TODO: A potentially significant performance optimization could be, if we have a LibcWrangler
# which is not "unknown", we don't need to clear the cache on objfile events (but probably still
# should on start events).


@pwndbg.lib.cache.cache_until("start", "objfile")
def __get_libc() -> tuple[Path, Path, LibcProvider]:
    """
    Find the active libc implementation and the associated libc and ld mappings.

    The process must be alive when this is called.

    If the program is statically linked, will return the main executable module's
    Path for the libc and ld path, and still try to infer the libc implementation.

    If no libc verifies anything, but there is at least one libc OR ld candidate
    mapping, it/they will be returned along with the "unknown" libc implementation.

    Returns:
        A tuple (libc mapping path, ld mapping path, libc implementation). Both of
            the returned Path's are resolved (absolute, followed symlinks).

    Raises:
        LibcNotFound: If the binary is dynamically linked and we couldn't find
            any candidate mappings.
    """
    # This function works by finding likely libc and ld mappings based on their
    # path names, and quering the libc implementations on them to see if any
    # claim the mapping as theirs. If noone claims the mappings, we return the
    # "unknown" libc implementation with the likely mappings.

    inf = pwndbg.dbg.selected_inferior()
    assert inf.alive()

    seen: set[str] = set()

    # Skip the executable
    maybe_main_module = inf.main_module_name()
    if maybe_main_module is not None:
        maybe_main_module = pwndbg.lib.path.clean_path(maybe_main_module)
        seen.add(maybe_main_module)

    all_sections: list[tuple[int, int, str, str]] = inf.module_section_locations()
    all_module_names: list[str] = [sec[3] for sec in all_sections]

    exact_libc_basename_matches: list[str] = [
        # glibc
        "libc.so.6",
        # musl and bionic (android)
        "libc.so",
    ]
    exact_ld_basename_matches: list[str] = [
        # x86_64 glibc
        "ld-linux-x86-64.so.2",
        # Common in CTF's
        "ld-linux.so",
        # x86_64 musl ld
        "ld-musl-x86_64.so.1",
    ]

    possible_libc_paths: list[str] = []
    possible_ld_paths: list[str] = []
    certain_libc_path: str | None = None
    certain_ld_path: str | None = None

    for path in all_module_names:
        if path in seen:
            continue
        seen.add(path)

        basename = os.path.basename(
            # Strip "target:" prefix used for remote debugging
            path[7:] if path.startswith("target:") else path
        )

        # Get absolute path and resolve symlinks if it seems plausible.
        # See #3641.
        path = pwndbg.lib.path.clean_path(path)

        # Check for libc
        if certain_libc_path is None and basename in exact_libc_basename_matches:
            # This is exceedingly likely to be the correct module.
            certain_libc_path = path
        elif libc_regex.search(basename) is not None:
            # Maybe the user loaded the libc with LD_PRELOAD.
            # Some common libc names: libc-2.36.so, libc6_2.36-0ubuntu4_amd64.so, libc.so
            possible_libc_paths.append(path)

        # Check for ld
        if certain_ld_path is None and basename in exact_ld_basename_matches:
            # This is exceedingly likely to be the correct module.
            certain_ld_path = path
        elif ld_regex.search(basename) is not None:
            possible_ld_paths.append(path)

    # Put the likeliest paths in the front. But also check the other ones
    # in case something else gets verified.
    # Though this would be extremely weird. Maybe we shouldn't allow it?
    if certain_libc_path:
        possible_libc_paths = [certain_libc_path] + possible_libc_paths
    if certain_ld_path:
        possible_ld_paths = [certain_ld_path] + possible_ld_paths

    # If we are statically linked, pass in the main module as it will contain
    # some libc stuff inside it (only the stuff that is actaully used).
    if not pwndbg.dbg.selected_inferior().is_dynamically_linked():
        # maybe_main_module should be non-None if the process is alive.
        assert maybe_main_module is not None
        possible_libc_paths = [maybe_main_module]
        possible_ld_paths = [maybe_main_module]

    # Let's see if any libc implementation verifies any of the
    # candidate paths we found.
    verified: tuple[str | None, str | None, LibcProvider | None] = __check_candidates(
        possible_libc_paths, possible_ld_paths
    )
    verified_libc_path, verified_ld_path, verified_libc_impl = verified

    if verified_libc_impl is not None:
        # Someone approved something!
        if verified_libc_path is not None and verified_ld_path is not None:
            return (Path(verified_libc_path), Path(verified_ld_path), verified_libc_impl)
        if verified_libc_path is not None:
            # We didn't get an approved ld path.
            # Lets ask the libc if it wants us to return the same path for the ld as for the libc,
            # or try to get an ld candidate.
            if not verified_libc_impl.libc_same_as_ld() and possible_ld_paths:
                ld_ret = possible_ld_paths[0]
            else:
                ld_ret = verified_libc_path
            return (Path(verified_libc_path), Path(ld_ret), verified_libc_impl)
        assert verified_ld_path is not None
        # We didn't get an approved libc path.
        # Lets ask the libc if it wants us to return the same path for the libc as for the ld,
        # or try to get an libc candidate.
        if not verified_libc_impl.libc_same_as_ld() and possible_libc_paths:
            libc_ret = possible_libc_paths[0]
        else:
            libc_ret = verified_ld_path
        return (Path(verified_ld_path), Path(libc_ret), verified_libc_impl)

    # Noone approved anything. If we have any candidate paths return them, otherwise raise exception.
    if possible_libc_paths and possible_ld_paths:
        return (Path(possible_libc_paths[0]), Path(possible_ld_paths[0]), unknown)
    if possible_libc_paths:
        return (Path(possible_libc_paths[0]), Path(possible_libc_paths[0]), unknown)
    if possible_ld_paths:
        return (Path(possible_ld_paths[0]), Path(possible_ld_paths[0]), unknown)
    # NOTE: We could also try to verify all of the other mappings in the address space, which would
    # sometimes yield us correct detection if the libc is very weirdly named, but it might be rare
    # enough and slow enough that it's not worth it. Not sure.
    # But if none of those get approved, we shouldn't return the first "candidate" match but really
    # raise.
    raise LibcNotFound("No candidate libc or ld mappings found.")


@pwndbg.lib.cache.cache_until("start", "objfile")
def get_libc() -> LibcProvider:
    _, _, libc = __get_libc()
    return libc


def which() -> LibcType:
    libc: LibcProvider = get_libc()
    return libc.type()


# ======== Public API =========


def has_exported_symbols() -> bool:
    """
    Do we have exported library symbols (e.g. fscanf, read, write)?

    If the library is dynamically linked, they will always be there. If it is statically
    linked and stripped, they may be missing.
    """
    return util.has_exported_symbols(str(filepath()))


def has_internal_symbols() -> bool:
    """
    Do we have internal library symbols?

    If the library is dynamically linked, even if it is stripped it will retain its
    exported symbols (e.g. fscanf) because they are required for dynamic linking.

    This funcions checks if the non-exported symbols (like __GI_exit, __run_exit_handlers,
    intitial) are also available.

    Symbols are global variables and functions. Internal symbols also come with debug info.
    """
    path, _, libc = __get_libc()
    return libc.has_internal_symbols(str(path))


def has_debug_info() -> bool:
    """
    Do we have debugging information like structure types?
    """
    libc: LibcProvider = get_libc()
    return libc.has_debug_info()


def filepath() -> Path:
    """
    The filepath of the libc shared object.

    There may not be a backing file for this Path if we are remote debugging.
    If the program is statically linked this will return the path of the main
    objfile.
    This may have the same value as loader_filepath() for some libc's.
    """
    path, _, _ = __get_libc()
    return path


def loader_filepath() -> Path:
    """
    The filepath of the ld shared object.

    There may not be a backing file for this Path if we are remote debugging.
    If the program is statically linked this will return the path of the main
    objfile.
    This may have the same value as filepath() for some libc's.
    """
    _, path, _ = __get_libc()
    return path


def addr() -> int:
    """
    The start load address of the libc shared object file.

    If the program is statically linked this will return the address of the main
    objfile.
    May be the same as loader_addr() for some libc's.
    """
    yes = pwndbg.aglib.vmmap.named_region_start(str(filepath()))
    # We know filepath() will return an actual mapped objfile, so
    # `yes` must be non-None.
    assert yes is not None
    return yes


def loader_addr() -> int:
    """
    The start load address of the ld shared object file.

    If the program is statically linked this will return the address of the main
    objfile.
    May be the same as addr() for some libc's.
    """
    yes = pwndbg.aglib.vmmap.named_region_start(str(loader_filepath()))
    # We know loader_filepath will return an actual mapped objfile, so
    # `yes` must be non-None.
    assert yes is not None
    return yes


def section_by_name(section_name: str) -> tuple[int, int, bytes] | None:
    """
    Returns pwndbg.aglib.elf.section_by_name() for the libc shared object file.
    """
    assert pwndbg.aglib.proc.alive()

    return pwndbg.aglib.elf.section_by_name(str(filepath()), section_name, try_local_path=True)


def section_address_by_name(section_name: str) -> int:
    """
    Get the start load address of the section `section_name` in the libc shared
    object file.
    """
    # TODO: If we are debugging a remote process, this might not work if GDB cannot load the so file
    libc_path: str = str(filepath())
    for (
        address,
        size,
        candidate_section_name,
        module_name,
    ) in pwndbg.dbg.selected_inferior().module_section_locations():
        if section_name == candidate_section_name and module_name == libc_path:
            return address
    return 0


def relocations_by_section_name(section_name: str) -> tuple[Relocation, ...]:
    """
    Returns pwndbg.aglib.elf.relocations_by_section_name() for the libc shared object file.
    """
    assert pwndbg.aglib.proc.alive()

    return pwndbg.aglib.elf.relocations_by_section_name(
        str(filepath()), section_name, try_local_path=True
    )


def urls() -> LibcURLs:
    """
    Get useful URLs regarding this libc implementation.
    """
    libc: LibcProvider = get_libc()
    try:
        ver = version()
        return libc.urls(ver)
    except NotImplementedError:
        return libc.urls(None)


def version() -> tuple[int, ...]:
    """
    Get the version of the libc implementation as a tuple.

    If you are calling this, you must know exactly which libc is being used.

    If the version couldn't be determined, (-1, -1) will be returned.
    """
    path, _, libc = __get_libc()
    return libc.version(str(path))


# ======== End of Public API =========



from __future__ import annotations

import pwndbg.aglib
import pwndbg.aglib.nearpc
import pwndbg.color.context as ctx_color
from pwndbg.aglib.disasm.instruction import ALL_JUMP_GROUPS
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.color import ColorConfig
from pwndbg.color import ColorParamSpec
from pwndbg.color import ljust_colored
from pwndbg.color import strip
from pwndbg.color import theme
from pwndbg.color.message import off
from pwndbg.color.message import on

c = ColorConfig(
    "disasm",
    [
        ColorParamSpec("branch", "bold", "color for disasm (branch/call instruction)"),
    ],
)

config_branch_on = theme.add_param(
    "disasm-branch-on", "✔", "marker for branches that WILL be taken"
)
config_branch_off = theme.add_param(
    "disasm-branch-off", "✘", "marker for branches that will NOT be taken"
)


# Returns colorized instructions assembly and operands, and checkmark if branch is taken
#  Example: `✔ je     _IO_file_xsputn+341`. Inline symbol replacements made. No annotation or branch targets shown.
def one_instruction(ins: PwndbgInstruction) -> str:
    asm = ins.asm_string

    # Highlight the current line if enabled
    if pwndbg.config.highlight_pc and ins.address == pwndbg.aglib.regs.pc:
        asm = ctx_color.highlight(asm)

    is_call_or_jump = ins.groups & ALL_JUMP_GROUPS

    # Style the instruction mnemonic if it's a call/jump instruction.
    if is_call_or_jump:
        asm = asm.replace(ins.mnemonic, c.branch(ins.mnemonic), 1)

    # If we know the conditional is taken, mark it as taken.
    if ins.condition == InstructionCondition.TRUE or ins.is_conditional_jump_taken:
        asm = on(f"{config_branch_on} ") + asm
    elif ins.condition == InstructionCondition.FALSE:
        asm = off(f"{config_branch_off} ") + asm
    else:
        asm = f"  {asm}"

    return asm


MIN_SPACING = 5
WHITESPACE_LIMIT = 20


# To making the padding visually nicer, the following padding scheme is used for annotations:
# All instructions in a group will have the same amount of left-adjusting spaces, so they are aligned.
# A group is defined as a sequence of instructions surrounded by instructions that can change the instruction pointer.
def instructions_and_padding(instructions: list[PwndbgInstruction]) -> list[str]:
    result: list[str] = []

    cur_padding_len = None

    # Stores intermediate padding results so we can do a final pass to clean up edges and jagged parts
    # None if padding doesn't apply to the instruction
    paddings: list[int | None] = []

    # Used for padding. List of groups.
    # Each group is a list of index into paddings list
    groups: list[list[int]] = []

    current_group: list[int] = []

    for i, (ins, asm) in enumerate(zip(instructions, (one_instruction(i) for i in instructions))):
        if ins.has_jump_target:
            sym = ins.target_string

            asm = f"{ljust_colored(asm, 36)} <{sym}>"

            paddings.append(None)
            if current_group:
                groups.append(current_group)
                current_group = []
        else:
            if ins.syscall is not None:
                asm += f" <{pwndbg.aglib.nearpc.c.syscall_name('SYS_' + ins.syscall_name)}>"

            # Padding the string for a nicer output
            # This path calculates the padding for each instruction - even if there we don't have annotations for it.
            # This allows groups to have uniform padding, even if some of the instructions don't have annotations
            current_group.append(i)

            raw_len = len(strip(asm))

            if cur_padding_len is None:
                cur_padding_len = raw_len + MIN_SPACING
            elif cur_padding_len - raw_len < MIN_SPACING:
                # Annotations are getting too close to the disasm, push them to the right again
                cur_padding_len = raw_len + MIN_SPACING
            else:
                # This path allows the padding to be smaller again
                # If the instruction has too much whitespace, put the annotation more to the left
                # Make sure there is an instruction after this one, and it's not a branch. Otherwise, maintain current indentation.
                if (
                    i < len(instructions) - 1
                    and not instructions[i + 1].has_jump_target
                    and cur_padding_len - raw_len > WHITESPACE_LIMIT
                ):
                    cur_padding_len = raw_len + MIN_SPACING

            # Give the padding to the instruction, so we can reuse it in the future
            if ins.annotation:
                if ins.annotation_padding is not None:
                    cur_padding_len = ins.annotation_padding
                else:
                    ins.annotation_padding = cur_padding_len

            paddings.append(cur_padding_len)

        result.append(asm)

    if current_group:
        groups.append(current_group)

    # Make instructions in each group aligned uniformly
    for group in groups:
        if len(group) == 0:
            continue

        # Find minimum spacing
        min_padding = max(paddings[index] for index in group)

        # Make all the paddings in this group have the same padding
        for index in group:
            paddings[index] = min_padding

    final_result = []

    # Final pass to apply final paddings to make alignment of blocks of instructions cleaner
    for i, (ins, asm, padding) in enumerate(zip(instructions, result, paddings)):
        # Padding being None implies a jump target - this is already baked into "asm"
        if ins.annotation and padding is not None:
            asm = f"{ljust_colored(asm, padding)}{ins.annotation}"

        final_result.append(asm)

    return final_result



from __future__ import annotations

import argparse

import pwndbg.commands
import pwndbg.lib.cache
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""
Toggles memoization (caching).

Useful for diagnosing caching-related bugs. Decreases performance.
""",
)


@pwndbg.commands.Command(parser, category=CommandCategory.PWNDBG)
def memoize() -> None:
    pwndbg.lib.cache.IS_CACHING = not pwndbg.lib.cache.IS_CACHING

    status = message.off("OFF (pwndbg will work slower, use only for debugging pwndbg)")
    if pwndbg.lib.cache.IS_CACHING:
        status = message.on("ON")

    print(f"Caching is now {status}")



"""
A few helpers for making things print pretty-like.
"""

from __future__ import annotations

import os
import sys

import pwndbg.color.context as ctx_color
import pwndbg.lib.config
from pwndbg import config
from pwndbg.color import ljust_colored
from pwndbg.color import rjust_colored
from pwndbg.color import strip
from pwndbg.color import theme

theme.add_param("banner-separator", "─", "repeated banner separator character")
theme.add_param("banner-title-surrounding-left", "[ ", "banner title surrounding char (left side)")
theme.add_param(
    "banner-title-surrounding-right", " ]", "banner title surrounding char (right side)"
)
title_position = theme.add_param(
    "banner-title-position",
    "center",
    "banner title position",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["center", "left", "right"],
)


def banner(title, target=sys.stdout, width=None, extra=""):
    title = title.upper()
    if width is None:
        _height, width = get_window_size(target)
    if title:
        title = f"{config.banner_title_surrounding_left}{ctx_color.banner_title(title)}{extra}{config.banner_title_surrounding_right}"
    if "left" == title_position:
        banner = ljust_colored(title, width, str(config.banner_separator))
    elif "right" == title_position:
        banner = rjust_colored(title, width, str(config.banner_separator))
    else:
        banner = rjust_colored(
            title, (width + len(strip(title))) // 2, str(config.banner_separator)
        )
        banner = ljust_colored(banner, width, str(config.banner_separator))
    return ctx_color.banner(banner)


def addrsz(address) -> str:
    return pwndbg.dbg.addrsz(address)


def get_window_size(target=sys.stdout) -> tuple[int, int]:
    fallback = (int(os.environ.get("LINES", 24)), int(os.environ.get("COLUMNS", 80)))
    if not target.isatty():
        return fallback
    if os.environ.get("PWNDBG_IN_TEST") is not None:
        return fallback

    if target in (sys.stdout, sys.stdin):
        # We can ask the debugger for the window size
        rows, cols = get_cmd_window_size()
        if rows is not None and cols is not None:
            return rows, cols

    try:
        term = os.get_terminal_size(target.fileno())
        return term.lines or fallback[0], term.columns or fallback[1]
    except Exception:
        pass

    return fallback


def get_cmd_window_size():
    return pwndbg.dbg.get_cmd_window_size()



from __future__ import annotations

import argparse

import gdb

import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.auxv
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory


@pwndbg.commands.Command("Gets the current file.", category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def getfile() -> None:
    print(repr(pwndbg.auxv.get().AT_EXECFN))


parser = argparse.ArgumentParser(description="Continue execution until an address or expression.")
parser.add_argument("target", type=int, help="Location to stop execution at")


@pwndbg.commands.Command(parser, category=CommandCategory.NEXT)
def xuntil(target) -> None:
    try:
        addr = target

        if not pwndbg.aglib.memory.peek(addr):
            print(message.error(f"Invalid address {addr:#x}"))
            return

        spec = f"*{addr:#x}"
    except (TypeError, ValueError):
        # The following gdb command will throw an error if the symbol is not defined.
        try:
            gdb.execute(f"info address {target}", to_string=True, from_tty=False)
        except gdb.error:
            print(message.error(f"Unable to resolve {target}"))
            return
        spec = target

    gdb.Breakpoint(spec, temporary=True)
    if pwndbg.aglib.proc.alive():
        gdb.execute("continue", from_tty=False)
    else:
        gdb.execute("run", from_tty=False)



"""
Emulation assistance from Unicorn.
"""

from __future__ import annotations

import binascii
import re
import string
from typing import NamedTuple

import capstone6pwndbg as C
import unicorn as U
import unicorn.ppc_const

import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.memory
import pwndbg.aglib.strings
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.enhance as E
import pwndbg.color.memory as mem_color
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.enhance
import pwndbg.lib.memory
import pwndbg.lib.regs
from pwndbg import color
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.color.syntax_highlight import syntax_highlight
from pwndbg.lib.arch import PWNDBG_SUPPORTED_ARCHITECTURES_TYPE

if pwndbg.dbg.is_gdblib_available():
    import gdb


def parse_consts(u_consts) -> dict[str, int]:
    """
    Unicorn "consts" is a python module consisting of a variable definition
    for each known entity. We repack it here as a dict for performance.

    Maps "UC_*" -> integer value of the constant
    """
    consts: dict[str, int] = {}
    for name in dir(u_consts):
        if name.startswith("UC_"):
            consts[name] = getattr(u_consts, name)
    return consts


# Generate Map<Register name, unicorn constant>
def create_reg_to_const_map(
    base_consts: dict[str, int], additional_mapping: dict[str, int] = None
) -> dict[str, int]:
    # base_consts is Map<"UC_*_REG_", constant>
    # additional mapping is the manually additions that add to the returned dict

    # Create a map of "register_name" -> Capstone ID, for faster lookup
    # Example of one field in the mapping for x86: { "RAX": 35 }
    reg_to_const: dict[str, int] = {}

    r = re.compile(r"^UC_.*_REG_(.*)$")
    for k, v in base_consts.items():
        # Use regex to match the Capstone register names to our register names.
        # Ex: extract "RCX" from "UC_X86_REG_RCX"
        # All are uppercase
        m = r.match(k)

        if m:
            reg_to_const[m.group(1)] = v

    if additional_mapping is not None:
        reg_to_const.update(additional_mapping)

    return reg_to_const


# Map our internal architecture names onto Unicorn Engine's architecture types.
arch_to_UC: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, int] = {
    "i386": U.UC_ARCH_X86,
    "x86-64": U.UC_ARCH_X86,
    "mips": U.UC_ARCH_MIPS,
    # "sparc": U.UC_ARCH_SPARC,
    "arm": U.UC_ARCH_ARM,
    "armcm": U.UC_ARCH_ARM,
    "aarch64": U.UC_ARCH_ARM64,
    "powerpc": U.UC_ARCH_PPC,
    "rv32": U.UC_ARCH_RISCV,
    "rv64": U.UC_ARCH_RISCV,
    "s390x": U.UC_ARCH_S390X,
}

# Architecture specific maps: Map<"UC_*_REG_*",constant>
arch_to_UC_consts: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, dict[str, int]] = {
    "i386": parse_consts(U.x86_const),
    "x86-64": parse_consts(U.x86_const),
    "mips": parse_consts(U.mips_const),
    "sparc": parse_consts(U.sparc_const),
    "arm": parse_consts(U.arm_const),
    "armcm": parse_consts(U.arm_const),
    "aarch64": parse_consts(U.arm64_const),
    "powerpc": parse_consts(unicorn.ppc_const),
    "rv32": parse_consts(U.riscv_const),
    "rv64": parse_consts(U.riscv_const),
    "s390x": parse_consts(U.s390x_const),
}

# Architecture specific maps: Map<reg_name, Unicorn constant>
arch_to_reg_const_map: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, dict[str, int]] = {
    "i386": create_reg_to_const_map(arch_to_UC_consts["i386"]),
    "x86-64": create_reg_to_const_map(
        arch_to_UC_consts["x86-64"],
        {"FSBASE": U.x86_const.UC_X86_REG_FS_BASE, "GSBASE": U.x86_const.UC_X86_REG_GS_BASE},
    ),
    "mips": create_reg_to_const_map(arch_to_UC_consts["mips"]),
    "sparc": create_reg_to_const_map(arch_to_UC_consts["sparc"]),
    "arm": create_reg_to_const_map(arch_to_UC_consts["arm"]),
    "armcm": create_reg_to_const_map(arch_to_UC_consts["armcm"]),
    "aarch64": create_reg_to_const_map(
        arch_to_UC_consts["aarch64"], {"CPSR": U.arm64_const.UC_ARM64_REG_NZCV}
    ),
    "powerpc": create_reg_to_const_map(
        arch_to_UC_consts["powerpc"],
        {
            "R0": unicorn.ppc_const.UC_PPC_REG_0,
            "SP": unicorn.ppc_const.UC_PPC_REG_1,
            "R2": unicorn.ppc_const.UC_PPC_REG_2,
            "R3": unicorn.ppc_const.UC_PPC_REG_3,
            "R4": unicorn.ppc_const.UC_PPC_REG_4,
            "R5": unicorn.ppc_const.UC_PPC_REG_5,
            "R6": unicorn.ppc_const.UC_PPC_REG_6,
            "R7": unicorn.ppc_const.UC_PPC_REG_7,
            "R8": unicorn.ppc_const.UC_PPC_REG_8,
            "R9": unicorn.ppc_const.UC_PPC_REG_9,
            "R10": unicorn.ppc_const.UC_PPC_REG_10,
            "R11": unicorn.ppc_const.UC_PPC_REG_11,
            "R12": unicorn.ppc_const.UC_PPC_REG_12,
            "R13": unicorn.ppc_const.UC_PPC_REG_13,
            "R14": unicorn.ppc_const.UC_PPC_REG_14,
            "R15": unicorn.ppc_const.UC_PPC_REG_15,
            "R16": unicorn.ppc_const.UC_PPC_REG_16,
            "R17": unicorn.ppc_const.UC_PPC_REG_17,
            "R18": unicorn.ppc_const.UC_PPC_REG_18,
            "R19": unicorn.ppc_const.UC_PPC_REG_19,
            "R20": unicorn.ppc_const.UC_PPC_REG_20,
            "R21": unicorn.ppc_const.UC_PPC_REG_21,
            "R22": unicorn.ppc_const.UC_PPC_REG_22,
            "R23": unicorn.ppc_const.UC_PPC_REG_23,
            "R24": unicorn.ppc_const.UC_PPC_REG_24,
            "R25": unicorn.ppc_const.UC_PPC_REG_25,
            "R26": unicorn.ppc_const.UC_PPC_REG_26,
            "R27": unicorn.ppc_const.UC_PPC_REG_27,
            "R28": unicorn.ppc_const.UC_PPC_REG_28,
            "R29": unicorn.ppc_const.UC_PPC_REG_29,
            "R30": unicorn.ppc_const.UC_PPC_REG_30,
            "R31": unicorn.ppc_const.UC_PPC_REG_31,
        },
    ),
    "rv32": create_reg_to_const_map(arch_to_UC_consts["rv32"]),
    "rv64": create_reg_to_const_map(arch_to_UC_consts["rv64"]),
    "s390x": create_reg_to_const_map(arch_to_UC_consts["s390x"]),
}

# Architectures for which we want to enable virtual TLB mode
enable_virtual_tlb: dict[PWNDBG_SUPPORTED_ARCHITECTURES_TYPE, bool] = {
    "s390x": True,
    "powerpc": True,
    "sparc": True,
}

# combine the flags with | operator. -1 for all
(
    NO_DEBUG,
    DEBUG_INIT,
    DEBUG_EXECUTING,
    DEBUG_MEM_MAP,
    DEBUG_HOOK_CHANGE,
    DEBUG_MEM_READ,
    DEBUG_EMU_START_STOP,
    DEBUG_INTERRUPT,
    DEBUG_TRACE,
) = (0, 1, 2, 4, 8, 16, 32, 64, 128)

DEBUG = NO_DEBUG
# DEBUG = -1 # ALL
# DEBUG = DEBUG_EXECUTING | DEBUG_MEM_MAP | DEBUG_MEM_READ

if DEBUG != NO_DEBUG:

    def debug(debug_type, fmt, args=()) -> None:
        if DEBUG & debug_type:
            print(fmt % args)

else:

    def debug(debug_type, fmt, args=()) -> None:
        pass


# Until Unicorn Engine provides full information about the specific instruction
# being executed for all architectures, we must rely on Capstone to provide
# that information.
arch_to_SYSCALL = {
    U.UC_ARCH_X86: [
        C.x86_const.X86_INS_SYSCALL,
        C.x86_const.X86_INS_SYSENTER,
        C.x86_const.X86_INS_SYSEXIT,
        C.x86_const.X86_INS_SYSRET,
        C.x86_const.X86_INS_IRET,
        C.x86_const.X86_INS_IRETD,
        C.x86_const.X86_INS_IRETQ,
        C.x86_const.X86_INS_INT,
        C.x86_const.X86_INS_INT1,
        C.x86_const.X86_INS_INT3,
    ],
    U.UC_ARCH_MIPS: [C.mips_const.MIPS_INS_SYSCALL],
    U.UC_ARCH_SPARC: [C.sparc_const.SPARC_INS_T],
    U.UC_ARCH_ARM: [C.arm_const.ARM_INS_SVC],
    U.UC_ARCH_ARM64: [C.aarch64_const.AARCH64_INS_SVC],
    U.UC_ARCH_PPC: [C.ppc_const.PPC_INS_SC],
    U.UC_ARCH_RISCV: [C.riscv_const.RISCV_INS_ECALL],
}

ARM_BANNED_INSTRUCTIONS = {
    C.arm.ARM_INS_MRC,
    C.arm.ARM_INS_MRRC,
    C.arm.ARM_INS_MRC2,
    C.arm.ARM_INS_MRRC2,
}
# We stop emulation when hitting these instructions, since they depend on co-processors or other information
# unavailable to the emulator
BANNED_INSTRUCTIONS = {
    "mips": {C.mips.MIPS_INS_RDHWR, C.mips.MIPS_INS_ALIAS_RDHWR},
    "arm": ARM_BANNED_INSTRUCTIONS,
    "armcm": ARM_BANNED_INSTRUCTIONS,
    "aarch64": {C.aarch64.AARCH64_INS_MRS},
}

# https://github.com/unicorn-engine/unicorn/issues/550
blacklisted_regs = ["ip", "cs", "ds", "es", "fs", "gs", "ss"]

"""
e = pwndbg.emu.emulator.Emulator()
e.until_jump()
"""


class InstructionExecutedResult(NamedTuple):
    address: int
    size: int


# Instantiating an instance of `Emulator` will start an instance
# with a copy of the current processor state.
class Emulator:
    def __init__(self) -> None:
        self.arch = pwndbg.aglib.arch.name

        if self.arch not in arch_to_UC:
            raise NotImplementedError(f"Cannot emulate code for {self.arch}")

        # Mapping of Pwndbg register name to Unicorn constant for the register
        self.const_regs = arch_to_reg_const_map[self.arch]

        self.uc_mode = self.get_uc_mode()
        debug(DEBUG_INIT, "# Instantiating Unicorn for %s", self.arch)
        debug(DEBUG_INIT, "uc = U.Uc(%r, %r)", (arch_to_UC[self.arch], self.uc_mode))
        self.uc = U.Uc(arch_to_UC[self.arch], self.uc_mode)

        if enable_virtual_tlb.get(self.arch, False):
            debug(DEBUG_INIT, "# Setting TLB mode to virtual")
            self.uc.ctl_set_tlb_mode(U.UC_TLB_VIRTUAL)  # type: ignore[attr-defined]

        self.reg_set: pwndbg.lib.regs.RegisterSet = pwndbg.aglib.regs.current

        # Whether the emulator is allowed to emulate instructions
        # There are cases when the emulator is incorrect or we want to disable it for certain instruction types,
        # and so we can set this to False to indicate that we should not allow the emulator to continue to step
        self.valid = True

        # Jump tracking state
        self._prev = None
        self._prev_size = None
        self._curr = None

        # The address of the last successfully executed instruction using single_step
        self.last_pc = None

        # (address_successfully_executed, size_of_instruction)
        self.last_single_step_result = InstructionExecutedResult(None, None)

        # Initialize the register state
        for emu_reg in self.reg_set.emulated_regs_order:
            reg = emu_reg.name
            enum = self.get_reg_enum(reg)

            if not reg:
                debug(DEBUG_INIT, "# Could not set register %r", reg)
                continue

            if reg in blacklisted_regs:
                debug(DEBUG_INIT, "Skipping blacklisted register %r", reg)
                continue
            value = pwndbg.aglib.regs.read_reg(reg)
            if None in (enum, value):
                if reg not in blacklisted_regs:
                    debug(DEBUG_INIT, "# Could not set register %r", reg)
                continue

            # Most registers are initialized to zero.
            # However, some registers (CPSR on AArch64) do not default to zero, so we must explicitly set them to 0
            if not emu_reg.force_write and value == 0:
                continue

            name = f"U.x86_const.UC_X86_REG_{reg.upper()}"
            debug(DEBUG_INIT, "uc.reg_write(%(name)s, %(value)#x)", locals())
            self.uc.reg_write(enum, value)

        # Add a hook for unmapped memory
        self.hook_add(U.UC_HOOK_MEM_UNMAPPED, self.hook_mem_invalid)

        # Always stop executing as soon as there's an interrupt.
        self.hook_add(U.UC_HOOK_INTR, self.hook_intr)

        # Map in the page that $pc is on
        self.map_page(pwndbg.aglib.regs.pc)

        # Instruction tracing
        if DEBUG & DEBUG_TRACE:
            self.hook_add(U.UC_HOOK_CODE, self.trace_hook)

    @property
    def last_step_succeeded(self) -> bool:
        return None not in self.last_single_step_result

    def read_register(self, name: str):
        reg = self.get_reg_enum(name)

        if reg:
            return self.uc.reg_read(reg)

        return None
        # raise AttributeError(f"AttributeError: {self!r} object has no register {name!r}")

    def pc(self) -> int:
        return self.read_register(self.reg_set.pc)

    # Read size worth of memory, return None on error
    def read_memory(self, address: int, size: int) -> bytes | None:
        # Don't attempt if the address is not mapped on the host process
        if not pwndbg.aglib.vmmap.find(address):
            return None

        value = None
        try:
            # Raises UcError if failed
            # If the memory is not mapped, it will fail. It will not attempt to run the UC_HOOK_MEM_UNMAPPED hook
            # https://github.com/unicorn-engine/unicorn/blob/d4b92485b1a228fb003e1218e42f6c778c655809/uc.c#L569
            value = self.uc.mem_read(address, size)
        except U.unicorn.UcError as e:
            # Attempt to map the page manually and try again
            if e.errno == U.UC_ERR_READ_UNMAPPED:
                try:
                    first_page = pwndbg.lib.memory.page_align(address)
                    last_page_exclusive = pwndbg.lib.memory.page_align(
                        address + size + pwndbg.lib.memory.PAGE_SIZE
                    )

                    for page_addr in range(
                        first_page, last_page_exclusive, pwndbg.lib.memory.PAGE_SIZE
                    ):
                        if not (self.map_page(page_addr)):
                            return None

                    # Pages are mapped, try again
                    value = self.uc.mem_read(address, size)

                except U.unicorn.UcError:
                    debug(DEBUG_MEM_READ, "Emulator failed to read memory at %#x, %r", (address, e))

                    return None
            else:
                return None

        return bytes(value)

    # Recursively dereference memory, return list of addresses
    # read_size typically must be either 1, 2, 4, or 8. It dictates the size to read
    # Naturally, if it is less than the pointer size, then only one value would be telescoped
    def telescope(self, address: int, limit: int, read_size: int = None) -> list[int]:
        read_size = read_size if read_size is not None else pwndbg.aglib.arch.ptrsize

        result = [address]

        # This prevents a crash in `unpack_size` below with big (SIMD) memory reads
        if not read_size <= 8:
            return result

        for i in range(limit):
            if result.count(address) >= 2:
                break

            value = self.read_memory(address, read_size)
            if value is not None:
                # address = pwndbg.aglib.arch.unpack(value)
                address = pwndbg.aglib.arch.unpack_size(value, read_size)
                address &= pwndbg.aglib.arch.ptrmask
                result.append(address)
            else:
                break

        return result

    # Given an address, return a string like the one `pwndbg.chain.format` returns,
    # reading from the emulator memory
    def format_telescope(self, address: int, limit: int) -> str:
        address_list = self.telescope(address, limit)
        return self.format_telescope_list(address_list, limit)

    def format_telescope_list(
        self, chain: list[int], limit: int, enhance_string_len: int = None
    ) -> str:
        # Code is near identical to pwndbg.chain.format, but takes into account reading from
        # the emulator's memory when necessary
        arrow_left = pwndbg.chain.c.arrow(f" {pwndbg.chain.config_arrow_left} ")
        arrow_right = pwndbg.chain.c.arrow(f" {pwndbg.chain.config_arrow_right} ")

        # Colorize the chain
        rest = []
        for link in chain:
            symbol = pwndbg.aglib.symbol.resolve_addr(link) or None
            if symbol:
                symbol = f"{link:#x} ({symbol})"
            rest.append(mem_color.get(link, symbol))

        # If the dereference limit is zero, skip any enhancements.
        if limit == 0:
            return rest[0]

        # Otherwise replace last element with the enhanced information.
        rest = rest[:-1]

        # Enhance the last entry
        # If there are no pointers (e.g. eax = 0x41414141), then enhance it
        if len(chain) == 1:
            enhanced = self.telescope_enhance(
                chain[-1], code=True, enhance_string_len=enhance_string_len
            )
        elif len(chain) < limit + 1:
            enhanced = self.telescope_enhance(
                chain[-2], code=True, enhance_string_len=enhance_string_len
            )
        else:
            enhanced = pwndbg.chain.c.contiguous_marker(f"{pwndbg.chain.config_contiguous}")

        if len(chain) == 1:
            return enhanced

        return arrow_right.join(rest) + arrow_left + enhanced

    def telescope_enhance(self, value: int, code: bool = True, enhance_string_len: int = None):
        # Near identical to pwndbg.enhance.enhance, just read from emulator memory

        # Determine if its on a page - we do this in the real processes memory
        page = pwndbg.aglib.vmmap.find(value)
        can_read = True
        if not page or None is pwndbg.aglib.memory.peek(value):
            can_read = False

        if not can_read:
            return E.integer(pwndbg.enhance.int_str(value))

        instr = None
        exe = page and page.execute
        rwx = page and page.rwx

        # For the purpose of following pointers, don't display
        # anything on the stack or heap as 'code'
        if "[stack" in page.objfile or "[heap" in page.objfile:
            rwx = exe = False

        if exe:
            pwndbg_instr = pwndbg.aglib.disasm.disassembly.one_raw(value)
            if pwndbg_instr:
                instr = f"{pwndbg_instr.mnemonic} {pwndbg_instr.op_str}"
                if pwndbg.config.syntax_highlight:
                    instr = syntax_highlight(instr)

        # szval = pwndbg.aglib.strings.get(value) or None
        # Read from emulator memory
        szval = self.memory_read_string(value, max_string_len=enhance_string_len, max_read=None)
        szval0 = szval
        if szval:
            szval = E.string(repr(szval))

        # Fix for case when we can't read the end address anyway (#946)
        if value + pwndbg.aglib.arch.ptrsize > page.end:
            return E.integer(pwndbg.enhance.int_str(value))

        # Read from emulator memory
        # intval = int(pwndbg.aglib.memory.get_typed_pointer_value(pwndbg.aglib.typeinfo.pvoid, value))
        read_value = self.read_memory(value, pwndbg.aglib.arch.ptrsize)
        if read_value is not None:
            # intval = pwndbg.aglib.arch.unpack(read_value)
            intval = pwndbg.aglib.arch.unpack_size(read_value, pwndbg.aglib.arch.ptrsize)
        else:
            # This occurs when Unicorn fails to read the memory - which it shouldn't, as the
            # read_memory call will map the pages necessary, and this function assumes
            # that the pointer is a valid pointer (as it has already been telescoped)
            intval = 0

        intval0 = intval
        intval = E.integer(
            pwndbg.lib.pretty_print.int_to_string(intval & pwndbg.aglib.arch.ptrmask)
        )

        retval = []

        if not code:
            instr = None

        # If it's on the stack, don't display it as code in a chain.
        if instr and "[stack" in page.objfile:
            retval = [intval, szval]
        # If it's RWX but a small value, don't display it as code in a chain.
        elif instr and rwx and intval0 < 0x1000:
            retval = [intval, szval]
        # If it's an instruction and *not* RWX, display it unconditionally
        elif instr and exe:
            if not rwx:
                if szval:
                    retval = [instr, szval]
                else:
                    retval = [instr]
            else:
                retval = [instr, intval, szval]

        # Otherwise strings have preference
        elif szval:
            if len(szval0) < pwndbg.aglib.arch.ptrsize:
                retval = [intval, szval]
            else:
                retval = [szval]

        # And then integer
        else:
            return E.integer(pwndbg.enhance.int_str(intval0))

        retval_final: tuple[str] = tuple(filter(lambda x: x is not None, retval))

        if len(retval_final) == 0:
            return E.unknown("???")

        if len(retval_final) == 1:
            return retval_final[0]

        return retval_final[0] + E.comment(color.strip(f" /* {'; '.join(retval_final[1:])} */"))

    # Return None if cannot find str
    def memory_read_string(self, address: int, max_string_len=None, max_read=None) -> str | None:
        if max_string_len is None:
            max_string_len = pwndbg.aglib.strings.length

        if max_read is None:
            max_read = pwndbg.aglib.strings.length

        # Read string
        sz = self.read_memory(address, max_read)
        if sz is None:
            return None

        try:
            sz = sz[: sz.index(b"\x00")]
        except ValueError:
            return None

        sz = sz.decode("latin-1", "replace")

        if not sz or not all(s in string.printable for s in sz):
            return None

        if len(sz) < max_string_len or not max_string_len:
            return sz

        return sz[:max_string_len] + "..."

    def update_pc(self, pc=None) -> None:
        if pc is None:
            pc = pwndbg.aglib.regs.pc
        self.uc.reg_write(self.get_reg_enum(self.reg_set.pc), pc)

    def read_thumb_bit(self) -> int:
        """
        Return 0 or 1, representing the status of the Thumb bit in the current Arm architecture

        This reads from the emulator itself, meaning this can be read to determine a state
        transitions between non-Thumb and Thumb mode

        Return None if the Thumb bit is not relevent to the current architecture

        Mimics the `read_thumb_bit` function defined in aglib/arch_mod.py
        """
        if self.arch == "arm":
            if (cpsr := self.read_register("cpsr")) is not None:
                return (cpsr >> 5) & 1
        elif self.arch == "armcm":
            if (xpsr := self.read_register("xpsr")) is not None:
                return (xpsr >> 24) & 1
        return 0

    def get_uc_mode(self):
        """
        Retrieve the mode used by Unicorn for the current architecture.
        """
        arch = pwndbg.aglib.arch.name
        mode = 0

        if arch == "armcm":
            mode |= (
                (U.UC_MODE_MCLASS | U.UC_MODE_THUMB)
                if (pwndbg.aglib.regs.read_reg("xpsr") & (1 << 24))
                else U.UC_MODE_MCLASS
            )

        elif arch in ("arm", "aarch64"):
            mode |= (
                U.UC_MODE_THUMB
                if (pwndbg.aglib.regs.read_reg("cpsr") & (1 << 5))
                else U.UC_MODE_ARM
            )

        elif (
            arch == "mips"
            and pwndbg.dbg.is_gdblib_available()
            and "isa32r6" in gdb.newest_frame().architecture().name()
        ):
            mode |= U.UC_MODE_MIPS32R6
        elif arch == "sparc":
            mode |= {4: U.UC_MODE_SPARC32, 8: U.UC_MODE_SPARC64}[pwndbg.aglib.arch.ptrsize]
        elif arch == "s390x":
            pass  # fails with invalid mode error otherwise
        else:
            mode |= {4: U.UC_MODE_32, 8: U.UC_MODE_64}[pwndbg.aglib.arch.ptrsize]

        if pwndbg.aglib.arch.endian == "little":
            mode |= U.UC_MODE_LITTLE_ENDIAN
        else:
            mode |= U.UC_MODE_BIG_ENDIAN

        return mode

    def map_page(self, page) -> bool:
        page = pwndbg.lib.memory.page_align(page)
        size = pwndbg.lib.memory.PAGE_SIZE

        # TODO: for Sparc, page size must be 0x2000

        debug(DEBUG_MEM_MAP, "# Mapping %#x-%#x", (page, page + size))

        try:
            data = pwndbg.aglib.memory.read(page, size)
            data = bytes(data)
        except pwndbg.dbg_mod.Error:
            debug(DEBUG_MEM_MAP, "Could not map page %#x during emulation! [exception]", page)
            return False

        if not data:
            debug(DEBUG_MEM_MAP, "Could not map page %#x during emulation! [no data]", page)
            return False

        debug(DEBUG_MEM_MAP, "uc.mem_map(%(page)#x, %(size)#x)", locals())
        self.uc.mem_map(page, size)

        debug(DEBUG_MEM_MAP, "# Writing %#x bytes", len(data))
        debug(DEBUG_MEM_MAP, "uc.mem_write(%(page)#x, ...)", locals())
        self.uc.mem_write(page, data)

        return True

    def hook_mem_invalid(self, uc, access, address, size: int, value, user_data) -> bool:
        debug(DEBUG_MEM_MAP, "# Invalid access at %#x, attempting to map the page", address)

        # Page-align the start address
        start = pwndbg.lib.memory.page_align(address)
        size = pwndbg.lib.memory.page_size_align(address + size - start)
        stop = start + size

        # Map each page with the permissions that we think it has.
        for page in range(start, stop, pwndbg.lib.memory.PAGE_SIZE):
            if not self.map_page(page):
                return False

        # Demonstrate that it's mapped
        # data = binascii.hexlify(self.uc.mem_read(address, size))
        # debug("# Memory is mapped: %#x --> %r", (address, data))

        return True

    def hook_intr(self, uc, intno, user_data) -> None:
        """
        We never want to emulate through an interrupt.  Just stop.
        """
        debug(DEBUG_INTERRUPT, "Got an interrupt - %d", intno)
        self.valid = False
        self.uc.emu_stop()

    def get_reg_enum(self, reg: str) -> int | None:
        """
        Returns the Unicorn Emulator enum code for the named register.

        Also supports general registers like 'sp' and 'pc'.
        """
        if not self.reg_set:
            return None

        # Look up the Unicorn enum for an exact register ('eax', 'ebp', 'r0')
        # This does not handle aliases, such as "sp" or "pc"
        e = self.const_regs.get(reg.upper(), None)
        if e is not None:
            return e

        return None

    def hook_add(self, *a, **kw):
        rv = self.uc.hook_add(*a, **kw)
        debug(DEBUG_HOOK_CHANGE, "%r = uc.hook_add(*%r, **%r)", (rv, a, kw))
        return rv

    def hook_del(self, *a, **kw):
        debug(DEBUG_HOOK_CHANGE, "uc.hook_del(*%r, **%r)", (a, kw))
        return self.uc.hook_del(*a, **kw)

    # Can throw a UcError(status)
    def emu_start(self, *a, **kw):
        debug(DEBUG_EMU_START_STOP, "uc.emu_start(*%r, **%r)", (a, kw))
        return self.uc.emu_start(*a, **kw)

    def emu_stop(self, *a, **kw):
        debug(DEBUG_EMU_START_STOP, "uc.emu_stop(*%r, **%r)", (a, kw))
        return self.uc.emu_stop(*a, **kw)

    def emulate_with_hook(self, hook, count=512) -> None:
        ident = self.hook_add(U.UC_HOOK_CODE, hook)

        pc: int = self.pc()
        # Unicorn appears to disregard the UC_MODE_THUMB mode passed into the constructor, and instead
        # determines Thumb mode based on the PC that is passed to the `emu_start` function
        # https://github.com/unicorn-engine/unicorn/issues/391
        #
        # Because we single-step the emulator, we always have to read the Thumb bit from the emulator
        # and set the least significant bit of the PC to 1 if the bit is 1 in order to enable Thumb mode
        # for the execution of the next instruction. If this `emulate_with_hook` executes multiple instructions
        # which have Thumb mode transitions, Unicorn will internally handle them.
        pc |= self.read_thumb_bit()

        try:
            self.emu_start(pc, 0, count=count)
        finally:
            self.hook_del(ident)

    def mem_read(self, *a, **kw):
        debug(DEBUG_MEM_READ, "uc.mem_read(*%r, **%r)", (a, kw))
        return self.uc.mem_read(*a, **kw)

    def until_jump(self, pc: int = None):
        """
        Emulates instructions starting at the specified address until the
        program counter is set to an address which does not linearly follow
        the previously-emulated instruction.

        Arguments:
            pc: Address to start at.  If `None`, uses the current instruction.

        Return:
            Returns a tuple containing the address of the jump instruction,
            and its target in the format (address, target).

            If emulation is forced to stop (e.g., because of a syscall or
            invalid memory access) then address is the instruction which
            could not be emulated through, and target will be None.

        Notes:
            This routine does not consider 'call $+5'
        """
        if pc is not None:
            self.update_pc(pc)

        # Set up the state.  Resetting this each time means that we will not ever
        # stop on the *current* instruction.
        self._prev = None
        self._prev_size = None
        self._curr = None

        # Add the jump hook, start emulating, and remove the hook.
        self.emulate_with_hook(self.until_jump_hook_code)

        # We're done emulating
        return self._prev, self._curr

    def until_jump_hook_code(self, _uc, address, instruction_size: int, _user_data) -> None:
        # We have not emulated any instructions yet.
        if self._prev is None:
            pass

        # We have moved forward one linear instruction, no branch or the
        # branch target was the next instruction.
        elif self._prev + self._prev_size == address:
            pass

        # We have branched!
        # The previous instruction does not immediately precede this one.
        else:
            self._curr = address
            debug(DEBUG_EXECUTING, "%#x %#X --> %#x", (self._prev, self._prev_size, self._curr))
            self.emu_stop()
            return

        self._prev = address
        self._prev_size = instruction_size

    def until_call(self, pc=None):
        addr, target = self.until_jump(pc)

        while target and not pwndbg.aglib.disasm.disassembly.one_raw(addr).call_like:
            addr, target = self.until_jump(target)

        return addr, target

    def until_syscall(self, pc=None):
        """
        Emulates instructions starting at the specified address until the program
        counter points at a syscall instruction (int 0x80, svc, etc.).
        """
        self.until_syscall_address = None
        self.emulate_with_hook(self.until_syscall_hook_code)
        return (self.until_syscall_address, None)

    def until_syscall_hook_code(self, uc, address, size: int, user_data) -> None:
        data = binascii.hexlify(self.mem_read(address, size))
        debug(
            DEBUG_EXECUTING, "# Executing instruction at %(address)#x with bytes %(data)s", locals()
        )
        self.until_syscall_address = address

    def single_step(self, pc=None, instruction: PwndbgInstruction | None = None) -> tuple[int, int]:
        """Steps one instruction.

        Yields:
            Each iteration, yields a tuple of (address_just_executed, instruction_size).

            Returns (None, None) upon failure to execute the instruction
        """

        # If the emulator has been manually marked as invalid, we should no longer step it
        if not self.valid:
            return InstructionExecutedResult(None, None)

        self.last_single_step_result = InstructionExecutedResult(None, None)

        pc = pc or self.pc()

        if instruction is None:
            instruction = pwndbg.aglib.disasm.disassembly.one_raw(pc)

            # If we don't know how to disassemble, bail.
            if instruction is None:
                debug(DEBUG_EXECUTING, "Can't disassemble instruction at %#x", pc)
                return self.last_single_step_result

        if instruction.id in BANNED_INSTRUCTIONS.get(self.arch, {}):
            debug(DEBUG_EXECUTING, "Hit illegal instruction at %#x", pc)
            return self.last_single_step_result

        debug(
            DEBUG_EXECUTING,
            "# Instruction: attempting to single-step at %#x: %s %s",
            (pc, instruction.mnemonic, instruction.op_str),
        )

        try:
            self.single_step_hook_hit_count = 0
            self.emulate_with_hook(self.single_step_hook_code, count=1)
            if not self.valid:
                return InstructionExecutedResult(None, None)

            # If above call does not throw an Exception, we successfully executed the instruction
            self.last_pc = pc
            debug(DEBUG_EXECUTING, "Unicorn now at pc=%#x", self.pc())
        except U.unicorn.UcError:
            debug(DEBUG_EXECUTING, "Emulator failed to execute instruction")
            self.last_single_step_result = InstructionExecutedResult(None, None)

        return self.last_single_step_result

    def single_step_iter(self, pc=None):
        a = self.single_step(pc)

        while a:
            yield a
            a = self.single_step(pc)

    # Whenever Unicorn is "about to execute" an instruction, this hook is called
    # https://github.com/unicorn-engine/unicorn/issues/1434
    def single_step_hook_code(self, _uc, address: int, instruction_size: int, _user_data) -> None:
        # For whatever reason, the hook will hit twice on
        # unicorn >= 1.0.2rc4, but not on unicorn-1.0.2rc1~unicorn-1.0.2rc3,
        # So we use a counter to ensure the code run only once
        if self.single_step_hook_hit_count == 0:
            debug(DEBUG_EXECUTING, "# single_step: %#-8x", address)
            self.last_single_step_result = InstructionExecutedResult(address, instruction_size)
            self.single_step_hook_hit_count += 1

    # For debugging
    def dumpregs(self) -> None:
        for reg in (
            list(self.reg_set.retaddr)
            + list(self.reg_set.misc)
            + list(self.reg_set.common)
            + list(self.reg_set.flags)
        ):
            enum = self.get_reg_enum(reg)

            if not reg or enum is None:
                print(f"# Could not dump register {reg!r}")
                continue

            name = f"U.x86_const.UC_X86_REG_{reg.upper()}"
            value = self.uc.reg_read(enum)
            print(f"uc.reg_read({name}) ==> {value:x}")

    def trace_hook(self, _uc, address, instruction_size: int, _user_data) -> None:
        data = binascii.hexlify(self.mem_read(address, instruction_size))
        debug(DEBUG_TRACE, "# trace_hook: %#-8x %r", (address, data))

    def __repr__(self) -> str:
        return f"Valid: {self.valid}, PC: {self.pc():#x}"



from __future__ import annotations


def to_signed(unsigned: int, bit_width: int):
    """
    Returns the signed number associated with the two's-complement binary representation of `unsigned`
    """
    unsigned = unsigned & ((1 << bit_width) - 1)
    extract_bit = 1 << (bit_width - 1)
    return unsigned - ((unsigned & extract_bit) << 1)


def logical_shift_left(n: int, shift_amt: int, bit_width: int):
    return (n << shift_amt) & ((1 << bit_width) - 1)


def logical_shift_right(n: int, shift_amt: int, bit_width: int):
    """
    `n` is truncated to the width of `bit_width` before the operation takes place.
    """
    n = n & ((1 << bit_width) - 1)
    return n >> shift_amt


def rotate_right(n: int, shift_amt: int, bit_width: int):
    """
    `n` is truncated to the width of `bit_width` before the operation takes place.
    """
    n = n & ((1 << bit_width) - 1)
    return ((n >> shift_amt) | (n << (bit_width - shift_amt))) & ((1 << bit_width) - 1)


def arithmetic_shift_right(n: int, shift_amt: int, bit_width: int):
    """
    This returns the value represented by the two's-complement binary representation of the final result.
    This means the result could be negative (if the top bit of the input is negative)

    `n` is truncated to the width of `bit_width` before the operation takes place.
    """
    n = n & ((1 << bit_width) - 1)

    result = logical_shift_right(n, shift_amt, bit_width)

    sign_extension_mask = (1 << (bit_width - shift_amt)) - 1
    # Replicate the sign bit if it's set
    if n & (1 << (bit_width - 1)):
        result |= ~sign_extension_mask

    return result



"""
Dynamic linking interface.

This module provides an interface to analyze various aspects of dynamically
linked programs.

It also defines a hook that watches for changes to the link map communicated by
ld.so, and exposes an event that other parts of pwndbg can tap into, but one
that may have a somewhat obtuse beahvior, due to limitations in GDB. See
`r_debug_install_link_map_changed_hook` for more information.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.color.message as message
import pwndbg.dbg_mod
from pwndbg.lib.elftypes import constants as elf


def _r_debug() -> int | None:
    """
    The easiest entry point into the link map is through the debug structure
    provided by ld.so. It provides a convenient pointer into the head of the
    link map list[1], and can be found at the address that corresponds to the
    `_r_debug` symbol[2].

    [1]: https://elixir.bootlin.com/glibc/latest/source/elf/link.h#L45
    [2]: https://elixir.bootlin.com/glibc/glibc-2.38/source/elf/dl-debug-symbols.S#L30
    """

    return pwndbg.aglib.symbol.lookup_symbol_addr("_r_debug")


def is_dynamic() -> bool:
    """
    Returns whether the current inferior is dynamic.

    Not all programs are dynamically linked, or even need the dynamic loader at
    all. Since this module is entirely reliant on at least the presence of the
    dynamic loader, and really only makes sense for dynamic programs, it should
    not be used at all with programs that don't participate in dynamic linkage,
    or when there is a dynamic linker, but we have no way to talk to it.
    """
    return _r_debug() is not None


# Reference to our hook in the link map update breakpoint, if it is installed.
R_DEBUG_LINK_MAP_CHANGED_HOOK = None
R_DEBUG_LINK_MAP_CHANGED_LISTENERS: set[Callable[..., Any]] = set()


def r_debug_link_map_changed_hook() -> Callable[[pwndbg.dbg_mod.StopPoint], bool]:
    """
    Hook that gets activated whenever the link map changes.

    The r_debug structure, in addition to having a refence to the head of the
    link map, also has, in its ABI-stable part, a reference to an address that
    can have a breakpoint attached to it, such that whenever the contents of the
    link map change, that breakpoint will be triggered[1].

    We take advantage of that here, by installing our own breakpoint in that
    location, and watching for trigger events, so that we can notify other bits
    of pwndbg that the contents of the `link_map()` function will be different.

    [1]: https://elixir.bootlin.com/glibc/glibc-2.37/source/elf/link.h#L52
    """

    skip_this = True

    def hook(_bp: pwndbg.dbg_mod.StopPoint) -> bool:
        # Skip every other trigger, we only care about the completed link map
        # that is available after the library is loaded.
        nonlocal skip_this
        if not skip_this:
            # Clear the cache that is tied to link map updates, and signal all of
            # the interested parties that this event has occurred.
            for listener in R_DEBUG_LINK_MAP_CHANGED_LISTENERS:
                listener()
        skip_this = not skip_this

        return False

    return hook


# FIXME: Obviously, having consumers call this function is not ideal. We really
# should find a way to install this automatically at the correct time. What we
# want is something that can run some arbitrary Python code at the same point
# in the lifecycle of the inferior as the user would be put in if they were to
# run `stepi`.
def r_debug_install_link_map_changed_hook() -> None:
    """
    Installs the r_debug-based hook to the change event of the link map.

    This function is a bit tricky, because ideally we want it to be run as
    soon as possible, before even the dynamic linker runs, but after both it and
    the main binary have been mapped into the address space of the inferior.
    While doing this manually would be trivial - seeing as there is a command
    in GDB that gives the user control at the exact place we would like -, there
    does not seem to be a way of easily doing this from inside Python.

    Because of this, parts of the code that rely on the hook should try calling
    this function and firing their own listeners manually at least once.
    """
    # TODO: Currently, reacting to this event has some unpleasant side effects
    # on usability. Until this can be fixed, this hook is unavailable.
    print(
        message.warn("r_brk hook is disabled. r_debug_install_link_map_changed_hook() does nothing")
    )
    return

    global R_DEBUG_LINK_MAP_CHANGED_HOOK
    if R_DEBUG_LINK_MAP_CHANGED_HOOK is not None:
        return

    r_debug_address = _r_debug()
    if r_debug_address is None:
        print(message.warn("symbol _r_debug is missing, cannot install link map change hook"))
        return

    r_debug = CStruct.r_debug()
    r_brk = r_debug.read(r_debug_address, "r_brk")

    bp = pwndbg.dbg.selected_inferior().break_at(
        pwndbg.dbg_mod.BreakpointLocation(r_brk),
        stop_hook=r_debug_link_map_changed_hook(),
        internal=True,
    )

    R_DEBUG_LINK_MAP_CHANGED_HOOK = bp


def r_debug_link_map_changed_add_listener(handler: Callable[..., Any]) -> None:
    """
    Install a callback to be called whenever r_debug signal of there being a
    change in the link map link map is triggered.

    Keep in mind this function may be called before the hook that calls the
    listeners is installed, and, until it is installed, no listener callbacks
    will actually be triggered. See `r_debug_install_link_map_changed_hook`.
    """
    R_DEBUG_LINK_MAP_CHANGED_LISTENERS.add(handler)


def r_debug_link_map_changed_remove_listener(handler: Callable[..., Any]) -> None:
    """
    Removes a listener previously installed with
    r_debug_link_map_changed_add_listener().
    """
    R_DEBUG_LINK_MAP_CHANGED_LISTENERS.remove(handler)


def link_map_head():
    """
    Acquires a reference to the head entry of the link map.
    """
    r_debug_address = _r_debug()
    if r_debug_address is None:
        print(message.warn("symbol _r_debug is missing, cannot find link map"))
        return None

    r_debug = CStruct.r_debug()

    r_map = r_debug.read(r_debug_address, "r_map")

    if r_map != 0:
        return LinkMapEntry(r_map)


def link_map():
    """
    Iterator over all the entries in the link map.
    """
    head = link_map_head()
    while head is not None:
        yield head
        head = head.next()


class LinkMapEntry:
    """
    An entry in the link map.
    """

    def __init__(self, address) -> None:
        self.link_map = CStruct.link_map()
        self.link_map_address = address

    def name(self):
        """
        The name of the binary image this entry describes.
        """
        ptr = self.link_map.read(self.link_map_address, "l_name")
        return pwndbg.aglib.memory.string(ptr)

    def dynamic(self):
        """
        The pointer to the memory mapped dynamic segment of the binary image.
        """
        return self.link_map.read(self.link_map_address, "l_ld")

    def load_bias(self):
        """
        The difference between the addresses in the data structures of the
        binary image and the actual location of the data being pointed to by
        them in the address space of the inferior. This number will never be
        negative.

        Aditionally, for DYN images, such as PIE executables and shared
        libraries, this value is the same as the base load address of the image.

        The term "load bias" comes from the ELF binary format loading procedure
        in the Linux Kernel.
        """
        return self.link_map.read(self.link_map_address, "l_addr")

    def next(self):
        """
        The next entry in the chain, if any.
        """
        ptr = self.link_map.read(self.link_map_address, "l_next")
        if ptr == 0:
            return None
        return LinkMapEntry(ptr)

    def prev(self):
        """
        The previous entry in the chain, if any.
        """
        ptr = self.link_map.read(self.link_map_address, "l_prev")
        if ptr == 0:
            return None
        return LinkMapEntry(ptr)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} node={self.link_map_address:#x} name={self.name()} load_bias={self.load_bias():#x} dynamic={self.dynamic():#x}>"


# Normally, only one entry for each tag is allowed to be present in the dynamic
# array for us to consider the dynamic array to be well-formed. Tags in this
# set are allowed to appear multiple times.
DYNAMIC_SECTION_ALLOW_MULTIPLE = {elf.DT_NEEDED}

# The DynamicSegment class expects some tags to always be present to function
# correctly. In this set we list them explicitly. Code in that class is allowed
# to presume these tags are always present after __init__.
DYNAMIC_SECTION_REQUIRED_TAGS = {
    elf.DT_STRTAB,
    elf.DT_STRSZ,
    elf.DT_SYMTAB,
    elf.DT_SYMENT,
}


class DynamicSegment:
    """
    Parser for the DYNAMIC segment present in a binary image.
    """

    strtab_addr = 0
    strtab_size = 0

    symtab_addr = 0
    symtab_elem = None

    entries_by_tag: dict[Any, Any] = {}

    has_jmprel = False
    has_rela = False
    has_rel = False

    jmprel_addr = 0
    rela_addr = 0
    rel_addr = 0

    jmprel_elem = None
    rela_elem = None
    rel_elem = None

    jmprel_r_sym_fn = None
    jmprel_r_info_fn = None

    rela_r_sym_fn = None
    rela_r_info_fn = None

    rel_r_sym_fn = None
    rel_r_info_fn = None

    def __init__(self, address, load_bias) -> None:
        # Enumerate the ElfNN_Dyn entries.
        count = 0
        elf_dyn = CStruct.elfNN_dyn()

        while elf_dyn.read(address + count * elf_dyn.size, "d_tag") != elf.DT_NULL:
            count += 1

        # Set up the fields used by the methods we use next.
        self.entries = count
        self.address = address
        self.load_bias = load_bias
        self.elf_dyn = elf_dyn

        # Map the tags we want to find to their respective entires. We don't
        # allow for repeats on most tags as they should only appear once in a
        # well-formed dynamic segment.
        sections: dict[Any, Any] = {}
        for i in range(self.entries):
            tag = self.dyn_array_read(i, "d_tag")
            if tag in sections:
                if tag not in DYNAMIC_SECTION_ALLOW_MULTIPLE:
                    raise RuntimeError(f"tag {tag:#x} repeated")

                if isinstance(sections[tag], list):
                    sections[tag].append(i)
                else:
                    sections[tag] = [sections[tag], i]
            else:
                sections[tag] = i
        for tag in DYNAMIC_SECTION_REQUIRED_TAGS:
            if tag not in sections:
                raise RuntimeError(f"missing requried tag {tag:#x}")
        self.entries_by_tag = sections

        # Setup the string table reference.
        self.strtab_addr = self.dyn_array_read_tag_val(elf.DT_STRTAB)
        self.strtab_size = self.dyn_array_read_tag_val(elf.DT_STRSZ)

        # Find the address of the symbol table and determine the correct version
        # of the ElfNN_Sym structure to use for this table, based on the size
        # of the elements given by DT_SYMENT.
        self.symtab_addr = self.dyn_array_read_tag_val(elf.DT_SYMTAB)

        syment = self.dyn_array_read_tag_val(elf.DT_SYMENT)
        if syment == 16:
            self.symtab_elem = CStruct.elf32_sym()
        elif syment == 24:
            self.symtab_elem = CStruct.elf64_sym()
        else:
            raise RuntimeError(
                f"unsupported value {syment} for DT_SYMENT, expected either 16 (Elf32_Sym) or 24 (Elf64_Sym)"
            )

        # Check the relocation sections, and perform some sanity checks.
        self.has_jmprel = (
            elf.DT_JMPREL in sections and elf.DT_PLTREL in sections and elf.DT_PLTRELSZ in sections
        )
        self.has_rela = (
            elf.DT_RELA in sections and elf.DT_RELASZ in sections and elf.DT_RELAENT in sections
        )
        self.has_rel = (
            elf.DT_REL in sections and elf.DT_RELSZ in sections and elf.DT_RELENT in sections
        )

        # Create the CStructs for the entries in each of our relocation sections
        # and make sure that their size matches the value of their respective
        # dynamic array element size entry. Additionally, pick selectors for
        # r_sym and r_type based on the size of the element of each of the
        # relocation sections.
        if self.has_rela:
            self.rela_addr = self.dyn_array_read_tag_val(elf.DT_RELA)
            self.rela_elem = CStruct.elfNN_rela()
            relaent = self.dyn_array_read_tag_val(elf.DT_RELAENT)
            assert self.rela_elem.size == relaent

            if self.dyn_array_read_tag_val(elf.DT_RELASZ) % self.rela_elem.size != 0:
                raise RuntimeError("DT_RELASZ is not divisible by DT_RELAENT")

            if self.rela_elem.size == 12:
                self.rela_r_sym = elf32_r_sym
                self.rela_r_type = elf32_r_type
            elif self.rela_elem.size == 24:
                self.rela_r_sym = elf64_r_sym
                self.rela_r_type = elf64_r_type
            else:
                raise RuntimeError(
                    f"DT_RELAENT is {self.rela_elem}, expected 12 (ELF32) or 24 (ELF64)"
                )

        if self.has_rel:
            self.rel_addr = self.dyn_array_read_tag_val(elf.DT_REL)
            self.rel_elem = CStruct.elfNN_rel()
            relent = self.dyn_array_read_tag_val(elf.DT_RELENT)
            assert self.rel_elem.size == relent

            if self.dyn_array_read_tag_val(elf.DT_RELSZ) % self.rel_elem.size != 0:
                raise RuntimeError("DT_RELSZ is not divisible by DT_RELENT")

            if self.rel_elem.size == 8:
                self.rel_r_sym = elf32_r_sym
                self.rel_r_type = elf32_r_type
            elif self.rel_elem.size == 16:
                self.rel_r_sym = elf64_r_sym
                self.rel_r_type = elf64_r_type
            else:
                raise RuntimeError(
                    f"DT_RELENT is {self.rel_elem}, expected 8 (ELF32) or 16 (ELF64)"
                )

        if self.has_jmprel:
            self.jmprel_addr = self.dyn_array_read_tag_val(elf.DT_JMPREL)
            pltrel = self.dyn_array_read_tag_val(elf.DT_PLTREL)
            if pltrel == elf.DT_RELA:
                self.jmprel_elem = CStruct.elfNN_rela()
                if elf.DT_RELAENT not in sections:
                    raise RuntimeError("DT_PLTREL is DT_RELA, but missing DT_RELAENT")
                assert self.jmprel_elem.size == self.dyn_array_read_tag_val(elf.DT_RELAENT)

                if self.rela_elem.size == 12:
                    self.jmprel_r_sym = elf32_r_sym
                    self.jmprel_r_type = elf32_r_type
                elif self.rela_elem.size == 24:
                    self.jmprel_r_sym = elf64_r_sym
                    self.jmprel_r_type = elf64_r_type
                else:
                    raise RuntimeError(
                        f"DT_RELAENT is {self.rela_elem}, expected 12 (ELF32) or 24 (ELF64)"
                    )

            elif pltrel == elf.DT_REL:
                self.jmprel_elem = CStruct.elfNN_rel()
                if elf.DT_RELENT not in sections:
                    raise RuntimeError("DT_PLTREL is DT_REL, but missing DT_RELENT")
                assert self.jmprel_elem.size == self.dyn_array_read_tag_val(elf.DT_RELENT)

                if self.jmprel_elem.size == 8:
                    self.jmprel_r_sym = elf32_r_sym
                    self.jmprel_r_type = elf32_r_type
                elif self.jmprel_elem.size == 16:
                    self.jmprel_r_sym = elf64_r_sym
                    self.jmprel_r_type = elf64_r_type
                else:
                    raise RuntimeError(
                        f"DT_RELENT is {self.rel_elem}, expected 8 (ELF32) or 16 (ELF64)"
                    )

            if self.dyn_array_read_tag_val(elf.DT_PLTRELSZ) % self.jmprel_elem.size != 0:
                raise RuntimeError("DT_PLTRELSZ is not divisible by the element size")

    def jmprel_has_addend(self):
        """
        Returns whether the `r_addend` field is available in entries of JMPREL.
        """
        assert self.has_jmprel
        return self.dyn_array_read_tag_val(elf.DT_PLTREL) == elf.DT_RELA

    def rela_read(self, i, field):
        """
        Reads the requested field from the entry of the given index in RELA.
        """
        assert self.has_rela
        count = self.rela_entry_count()
        if i >= count:
            raise ValueError(f"tried to read entry {i} in RELA with only {count} entries")

        transform = lambda x: x
        if field == "r_sym":
            transform = self.rela_r_sym
            field = "r_info"
        elif field == "r_type":
            transform = self.rela_r_type
            field = "r_info"

        return transform(self.rela_elem.read(self.rela_addr + i * self.rela_elem.size, field))

    def rel_read(self, i, field):
        """
        Reads the requested field from the entry of the given index in REL.
        """
        assert self.has_rel
        count = self.rel_entry_count()
        if i >= count:
            raise ValueError(f"tried to read entry {i} in REL with only {count} entries")

        transform = lambda x: x
        if field == "r_sym":
            transform = self.rel_r_sym
            field = "r_info"
        elif field == "r_type":
            transform = self.rel_r_type
            field = "r_info"

        return transform(self.rel_elem.read(self.rel_addr + i * self.rel_elem.size, field))

    def jmprel_read(self, i, field):
        """
        Reads the requested field from the entry of the given index in JMPREL.
        """
        assert self.has_jmprel
        count = self.jmprel_entry_count()
        if i >= count:
            raise ValueError(f"tried to read entry {i} in JMPREL with only {count} entries")

        transform = lambda x: x
        if field == "r_sym":
            transform = self.jmprel_r_sym
            field = "r_info"
        elif field == "r_type":
            transform = self.jmprel_r_type
            field = "r_info"

        return transform(self.jmprel_elem.read(self.jmprel_addr + i * self.jmprel_elem.size, field))

    def rela_entry_count(self):
        """
        Returns the number of RELA entries.
        """
        assert self.has_rela
        relasz = self.dyn_array_read_tag_val(elf.DT_RELASZ)
        relaent = self.dyn_array_read_tag_val(elf.DT_RELAENT)

        return relasz // relaent

    def rel_entry_count(self):
        """
        Returns the number of REL entries.
        """
        assert self.has_rel
        relsz = self.dyn_array_read_tag_val(elf.DT_RELSZ)
        relent = self.dyn_array_read_tag_val(elf.DT_RELENT)

        return relsz // relent

    def jmprel_entry_count(self):
        """
        Returns the number of JMPREL entries.
        """
        assert self.has_jmprel
        pltrelsz = self.dyn_array_read_tag_val(elf.DT_PLTRELSZ)
        pltrelent = self.jmprel_elem.size

        return pltrelsz // pltrelent

    def string(self, i):
        """
        Reads the string at index i from the string table.
        """
        if i >= self.strtab_size:
            raise ValueError(
                f"tried to read entry {i} in string table with only {self.entries} bytes"
            )
        return pwndbg.aglib.memory.string(self.strtab_addr + i)

    def symtab_read(self, i, field):
        """
        Reads the requested field from the entry of given index in the symbol
        table.
        """
        return self.symtab_elem.read(self.symtab_addr + i * self.symtab_elem.size, field)

    def dyn_array_read(self, i, field):
        """
        Reads the requested field from the entry of given index in the dynamic
        array.
        """
        if i >= self.entries:
            raise ValueError(
                f"tried to read from entry {i} in dynamic array with only {self.entries} entries"
            )
        return self.elf_dyn.read(self.address + i * self.elf_dyn.size, field)

    def dyn_array_read_tag_val(self, tag):
        """
        Reads the `d_un` field from the entry of given tag in the dynamic
        array. Must not be a tag that allows multiple entries.
        """
        return self.dyn_array_read(self.entries_by_tag[tag], "d_un")


def elf32_r_sym(r_info):
    """
    Returns the r_sym portion of the r_info relocation field for ELF32.
    """
    return (r_info >> 8) & 0xFFFFFF


def elf32_r_type(r_info):
    """
    Returns the r_type portion of the r_info relocation field for ELF32.
    """
    return r_info & 0xFF


def elf64_r_sym(r_info):
    """
    Returns the r_sym portion of the r_info relocation field for ELF64.
    """
    return (r_info >> 32) & 0xFFFFFFFF


def elf64_r_type(r_info):
    """
    Returns the r_type portion of the r_info relocation field for ELF64.
    """
    return r_info & 0xFFFFFFFF


class CStruct:
    """
    Utility class for reading fields off of C structs.

    Without proper debug information it cannot be guaranteed that the calculated
    field offsets are correct, therefore, reasonable caution should be exercised
    when using this class. The assumptions made are:
     - Padding is added between fields so that all internal members are
       correctly aligned, as long as the struct itself is correctly aligned.
     - The alignment of the struct is the same as the alignment of its most
       strictly aligned member.
     - Padding is added to the end of the struct so that sequentially laid out
       instances are always correctly aligned.
     - Stuct sizes must be greater than or equal to 1 byte.

    While these assumptions do not apply in all cases, they should be good
    enough for the structs in ld.so and in the ELF program images.
    """

    types: dict[str, pwndbg.dbg_mod.Type] = {}
    offsets: dict[str, int] = {}
    converters: dict[str, type] = {}
    size = 0
    align = 0

    @staticmethod
    def link_map():
        """
        Creates a new instance describing the ABI-stable part of the link_map
        struct.
        """
        return CStruct(
            [
                ("l_addr", pwndbg.aglib.typeinfo.size_t, int),
                ("l_name", pwndbg.aglib.typeinfo.char.pointer(), int),
                ("l_ld", pwndbg.aglib.typeinfo.pvoid, int),
                ("l_next", pwndbg.aglib.typeinfo.pvoid, int),
                ("l_prev", pwndbg.aglib.typeinfo.pvoid, int),
            ]
        )

    @staticmethod
    def r_debug():
        """
        Creates a new instance describing the ABI-stable part of the r_debug
        struct.
        """
        return CStruct(
            [
                ("r_version", pwndbg.aglib.typeinfo.uint, int),
                ("r_map", pwndbg.aglib.typeinfo.pvoid, int),
                ("r_brk", pwndbg.aglib.typeinfo.pvoid, int),
            ]
        )

    @staticmethod
    def elfNN_dyn():
        """
        Creates a new instance describing the ElfNN_Dyn structure, suitable for
        the architecture of the inferior.
        """
        return CStruct(
            [
                ("d_tag", pwndbg.aglib.typeinfo.size_t, int),
                ("d_un", pwndbg.aglib.typeinfo.size_t, int),
            ]
        )

    @staticmethod
    def elfNN_rel():
        """
        Creates a new instance describing the ElfNN_Rel structure, suitable for
        the architecture of the inferior.
        """
        return CStruct(
            [
                ("r_offset", pwndbg.aglib.typeinfo.size_t, int),
                ("r_info", pwndbg.aglib.typeinfo.size_t, int),
            ]
        )

    @staticmethod
    def elfNN_rela():
        """
        Creates a new instance describing the ElfNN_Rela structure, suitable for
        the architecture of the inferior.
        """
        return CStruct(
            [
                ("r_offset", pwndbg.aglib.typeinfo.size_t, int),
                ("r_info", pwndbg.aglib.typeinfo.size_t, int),
                ("r_addend", pwndbg.aglib.typeinfo.size_t, int),
            ]
        )

    @staticmethod
    def elf32_sym():
        """
        Creates a new instance describing the Elf32_Sym srtucture.
        """
        # FIXME: ELF types have an exact size. We want our GDB types to match
        # whatever the platform's exact sized integer types are, but, because of
        # how these types are resolved, that might not always be the case.
        #
        # It's better to fail loudly here than to fail silently later.
        assert pwndbg.aglib.typeinfo.uint32.sizeof == 4
        assert pwndbg.aglib.typeinfo.uint16.sizeof == 2
        assert pwndbg.aglib.typeinfo.uint8.sizeof == 1

        return CStruct(
            [
                ("st_name", pwndbg.aglib.typeinfo.uint32, int),
                ("st_value", pwndbg.aglib.typeinfo.uint32, int),
                ("st_size", pwndbg.aglib.typeinfo.uint32, int),
                ("st_info", pwndbg.aglib.typeinfo.uint8, int),
                ("st_other", pwndbg.aglib.typeinfo.uint8, int),
                ("st_shndx", pwndbg.aglib.typeinfo.uint16, int),
            ]
        )

    @staticmethod
    def elf64_sym():
        """
        Creates a new instance describing the Elf64_Sym structure.
        """

        # FIXME: Same issue as elf32_sym()
        assert pwndbg.aglib.typeinfo.uint64.sizeof == 8
        assert pwndbg.aglib.typeinfo.uint32.sizeof == 4
        assert pwndbg.aglib.typeinfo.uint16.sizeof == 2
        assert pwndbg.aglib.typeinfo.uint8.sizeof == 1

        return CStruct(
            [
                ("st_name", pwndbg.aglib.typeinfo.uint32, int),
                ("st_info", pwndbg.aglib.typeinfo.uint8, int),
                ("st_other", pwndbg.aglib.typeinfo.uint8, int),
                ("st_shndx", pwndbg.aglib.typeinfo.uint16, int),
                ("st_value", pwndbg.aglib.typeinfo.uint64, int),
                ("st_size", pwndbg.aglib.typeinfo.uint64, int),
            ]
        )

    def __init__(self, fields: list[tuple[str, pwndbg.dbg_mod.Type, type]]) -> None:
        # Calculate the offset of all of the fields in the struct.
        current_offset = 0
        alignment = 1
        for entry in fields:
            name = entry[0]
            ty = entry[1]
            if len(entry) > 2:
                conv = entry[2]
            else:
                conv = None

            # Pad the offset so that the field is correctly aligned.
            if current_offset % ty.alignof != 0:
                current_offset += ty.alignof - (current_offset % ty.alignof)

            # Save the alignment requirements of the strictest element.
            if ty.alignof > alignment:
                alignment = ty.alignof

            self.offsets[name] = current_offset
            self.types[name] = ty
            self.converters[name] = conv

            current_offset += ty.sizeof

        # We don't consider zero-sized structures to be valid.
        if current_offset == 0:
            current_offset = 1

        # Pad the end of the struct so that the next instance of this struct in
        # an array is correctly aligned.
        if current_offset % alignment != 0:
            current_offset += alignment - (current_offset % alignment)

        self.size = current_offset
        self.align = alignment

    def read(self, address, name):
        """
        Reads the field with the given name from the struct instance located at
        the given address.
        """
        val = pwndbg.aglib.memory.get_typed_pointer_value(
            self.types[name], address + self.offsets[name]
        )
        if self.converters[name] is not None:
            return self.converters[name](val)
        return val

    def has_field(self, name) -> bool:
        """
        Returns whether a field with the given name exists in this struct.
        """
        return name in self.offsets



from __future__ import annotations

import functools
import re
from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from typing import TYPE_CHECKING
from typing import TypeVar

from elftools.elf.elffile import ELFFile
from typing_extensions import ParamSpec

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.structures
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.kernel.structs
import pwndbg.lib.memory
import pwndbg.search
from pwndbg.lib import Status
from pwndbg.lib import TypeNotFoundError
from pwndbg.lib import TypeNotRecoveredError
from pwndbg.lib.regs import BitFlags

if TYPE_CHECKING:
    import pwndbg.aglib.kernel.kconfig_mod
    import pwndbg.aglib.kernel.paging
    import pwndbg.aglib.kernel.slab
    import pwndbg.aglib.kernel.symbol
    import pwndbg.aglib.kernel.vmmap

_kconfig: pwndbg.aglib.kernel.kconfig_mod.Kconfig | None = None

P = ParamSpec("P")
D = TypeVar("D")
T = TypeVar("T")


def BIT(shift: int):
    assert 0 <= shift < 64
    return 1 << shift


def has_debug_symbols(*required: str, checkall: bool = True) -> bool:
    if not required:
        required = ("commit_creds",)
    required_syms_iter = (pwndbg.aglib.symbol.lookup_symbol(sym) is not None for sym in required)
    return all(required_syms_iter) if checkall else any(required_syms_iter)


@pwndbg.lib.cache.cache_until("objfile")
def has_debug_info() -> bool:
    path = pwndbg.aglib.proc.exe()
    if path is None:
        return False
    vmlinux = open(path, "rb")
    elf = ELFFile(vmlinux)
    return any(section.name == ".debug_info" for section in elf.iter_sections())


def requires_debug_symbols(
    *required: str, checkall=False, default: D = None
) -> Callable[[Callable[P, T]], Callable[P, T | D]]:
    def decorator(f: Callable[P, T]) -> Callable[P, T | D]:
        @functools.wraps(f)
        def func(*args: P.args, **kwargs: P.kwargs) -> T | D:
            if has_debug_symbols(*required, checkall=checkall):
                return f(*args, **kwargs)

            # If the user doesn't want an exception thrown when debug symbols are
            # not available, they can instead provide a default return value
            if default is not None:
                return default

            raise Exception(
                f"Function {f.__name__} requires {'all' if checkall else 'any'} of the following symbols: {required}"
            )

        return func

    return decorator


def requires_debug_info(default: D = None) -> Callable[[Callable[P, T]], Callable[P, T | D]]:
    def decorator(f: Callable[P, T]) -> Callable[P, T | D]:
        @functools.wraps(f)
        def func(*args: P.args, **kwargs: P.kwargs) -> T | D:
            if has_debug_info():
                return f(*args, **kwargs)

            # If the user doesn't want an exception thrown when debug symbols are
            # not available, they can instead provide a default return value
            if default is not None:
                return default

            raise Exception(f"Function {f.__name__} requires .debug_info section")

        return func

    return decorator


# Set by pwndbg.aglib.kernel.symbol.load_common_structs_on_load_linux() when page typeinfo
# recovery fails.
page_typeinfo_recovery_failure: None | TypeNotRecoveredError = None


def typeinfo_recovery(
    name: str, requires_kversion: bool = False, requires_kbase: bool = False
) -> Callable[[Callable[P, str]], Callable[P, None]]:
    def decorator(f: Callable[P, str]) -> Callable[P, None]:
        # returns true if the type exists or has been successfully recovered
        @functools.wraps(f)
        def func(*args: P.args, **kwargs: P.kwargs) -> None:
            if not pwndbg.dbg.selected_inferior().is_linux():
                # make sure the target is linux, should we specify symbols instead?
                raise TypeNotRecoveredError(name, "target is not linux")
            if has_debug_info():
                return
            if pwndbg.aglib.typeinfo.lookup_types(name) is not None:
                return
            if requires_kversion and kversion() is None:
                raise TypeNotRecoveredError(name, "kernel version is unavailable")
            if requires_kbase and kbase() is None:
                raise TypeNotRecoveredError(name, "kernel base not found")

            try:
                result = f(*args, **kwargs)
            except TypeNotFoundError as e:
                # typeinfo_recovery functions depend on
                # pwndbg.aglib.kernel.symbol.load_common_structs_on_load_linux()
                # succeeding and will try to directly read those types from the debbuger
                # like e.g. `pwndbg.aglib.memory.get_typed_pointer("struct list_head", db_list)`
                # This will raise a TypeNotFoundError exception.
                if page_typeinfo_recovery_failure is not None:
                    raise page_typeinfo_recovery_failure
                raise TypeNotRecoveredError(name, str(e))
            except AssertionError as e:
                # FIXME: Some type recovery functions `assert` under the assumption that the assert
                # will be caught here.
                raise TypeNotRecoveredError(name, str(e))

            fname = name.split()[-1] + "_structs"
            err: Status = pwndbg.aglib.structures.add(fname, result)
            if err.is_failure():
                raise TypeNotRecoveredError(name, err.message)
            return

        return func

    return decorator


@pwndbg.lib.cache.cache_until("start")
def nproc() -> int:
    """Returns the number of processing units available, similar to nproc(1)"""
    return len(pwndbg.dbg.selected_inferior().send_monitor("info cpus").splitlines())


@pwndbg.lib.cache.cache_until("stop")
def first_kernel_ro_page() -> pwndbg.lib.memory.Page | None:
    """Returns the first kernel mapping which contains the linux_banner"""
    base = kbase()
    if base is None:
        return None

    banner = pwndbg.aglib.symbol.lookup_symbol_addr("linux_banner")
    fallback_mappings = []
    for mapping in pwndbg.aglib.kernel.vmmap.kernel_vmmap_pages():
        if mapping.vaddr < base:
            continue
        if banner is not None and banner in mapping:
            return mapping
        if not mapping.read or mapping.write or mapping.execute:
            fallback_mappings.append(mapping)
            continue

        result = next(pwndbg.search.search(b"Linux version", mappings=[mapping]), None)

        if result:
            return mapping
    # optimization: observe that the first Linux kernel region is the kernel text so search it last
    # it now finds the first ro page almost instantly even for kernels that are partially initialized
    # should find it within the first few page chunks if debugging linux kernel (reason for [:10])
    for mapping in fallback_mappings[1:10] + [fallback_mappings[0]]:
        # this loop handles when the kernel has not finished initialization
        # and the permission of the first ro page has not been properly set
        result = next(pwndbg.search.search(b"Linux version", mappings=[mapping]), None)

        if result:
            return mapping

    return None


@pwndbg.lib.cache.cache_until("objfile")
def kconfig() -> pwndbg.aglib.kernel.kconfig_mod.Kconfig:
    global _kconfig
    config_start, config_end = None, None
    if has_debug_symbols():
        config_start = pwndbg.aglib.symbol.lookup_symbol_addr("kernel_config_data")
        config_end = pwndbg.aglib.symbol.lookup_symbol_addr("kernel_config_data_end")
    else:
        mapping = first_kernel_ro_page()
        result = None
        if mapping is not None:
            result = next(pwndbg.search.search(b"IKCFG_ST", mappings=[mapping]), None)

        if result is not None:
            config_start = result + len("IKCFG_ST")
            config_end = next(pwndbg.search.search(b"IKCFG_ED", start=config_start), None)
    if (
        not config_start
        or not config_end
        or not pwndbg.aglib.memory.is_kernel(config_start)
        or not pwndbg.aglib.memory.is_kernel(config_end)
        or config_start >= config_end
    ):
        _kconfig = pwndbg.aglib.kernel.kconfig_mod.Kconfig(None)
        return _kconfig

    config_size = config_end - config_start

    compressed_config = pwndbg.aglib.memory.read(config_start, config_size)
    _kconfig = pwndbg.aglib.kernel.kconfig_mod.Kconfig(compressed_config)
    return _kconfig


@requires_debug_symbols("saved_command_line", default="")
@pwndbg.lib.cache.cache_until("start")
def kcmdline() -> str:
    addr = pwndbg.aglib.symbol.lookup_symbol_addr("saved_command_line")
    if not addr:
        return ""
    cmdline_addr = pwndbg.aglib.memory.read_pointer_width(addr)
    return pwndbg.aglib.memory.string(cmdline_addr).decode("ascii")


@pwndbg.lib.cache.cache_until("start")
def kversion() -> str | None:
    try:
        if has_debug_symbols("linux_banner"):
            version_addr = pwndbg.aglib.symbol.lookup_symbol_addr("linux_banner")
            result = pwndbg.aglib.memory.string(version_addr).decode("ascii").strip()
            assert len(result) > 0
            return result
    except Exception:
        pass
    mapping = first_kernel_ro_page()
    if mapping is None:
        return None
    version_addr = next(pwndbg.search.search(b"Linux version", mappings=[mapping]), None)
    if version_addr is None:
        return None
    return pwndbg.aglib.memory.string(version_addr).decode("ascii").strip()


@pwndbg.lib.cache.cache_until("start")
def krelease() -> tuple[int, ...] | None:
    _kversion = kversion()
    if _kversion is None:
        return None
    match = re.search(r"Linux version (\d+)\.(\d+)(?:\.(\d+))?", _kversion)
    if match:
        return tuple(int(x) for x in match.groups() if x)
    raise Exception("Linux version tuple not found")


def get_idt_entries() -> list[pwndbg.lib.kernel.structs.IDTEntry]:
    """
    Retrieves the IDT entries from memory.
    """
    base = pwndbg.aglib.regs.idt
    limit = pwndbg.aglib.regs.idt_limit

    size = pwndbg.aglib.arch.ptrsize * 2
    num_entries = min((limit + 1) // size, 256)

    entries = []

    # TODO: read the entire IDT in one call?
    for i in range(num_entries):
        entry_addr = base + i * size
        entry = pwndbg.lib.kernel.structs.IDTEntry(pwndbg.aglib.memory.read(entry_addr, size))
        entries.append(entry)

    return entries


def current_cpu() -> int:
    return pwndbg.dbg.selected_thread().index() - 1


def get_double_linked_list(head: int, minlen: int = 0x1, maxlen: int = 0x1000) -> list[int] | None:
    # head is a pointer to the double linked list
    # None if not a doubly linked list
    if not pwndbg.aglib.memory.is_kernel(head):
        return None
    nxt = head
    result = []
    for _ in range(maxlen):
        if not pwndbg.aglib.memory.is_kernel(nxt):
            return None
        result.append(nxt)
        nxt = pwndbg.aglib.memory.read_pointer_width(nxt)
        if nxt == result[0]:
            break
    if nxt != result[0]:
        return None
    if len(result) < minlen:
        return None
    for i, nxt in enumerate(result):
        p = pwndbg.aglib.memory.read_pointer_width(nxt + pwndbg.aglib.arch.ptrsize)
        if p != result[i - 1]:
            return None
    return result


def in_kmem_cache(val: int, name: str, strict: bool = True) -> bool:
    # name is a substr of any of the target caches' names
    _, cache = pwndbg.aglib.kernel.slab.find_containing_slab_cache(val)
    if not cache:
        return False
    if strict:
        return name == cache.name
    return name in cache.name


class ArchOps(ABC):
    # More information on the physical memory model of the Linux kernel and
    # especially the mapping between pages and page frame numbers (pfn) can
    # be found at https://docs.kernel.org/mm/memory-model.html
    # The provided link also includes guidance on detecting the memory model in
    # use through kernel configuration, enabling support for additional models
    # in the page_to_pfn() and pfn_to_page() methods in the future.

    @abstractmethod
    def per_cpu(
        self, addr: int | pwndbg.dbg_mod.Value, cpu: int | None = None
    ) -> pwndbg.dbg_mod.Value | None:
        raise NotImplementedError()

    @abstractmethod
    def virt_to_phys(self, virt: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def phys_to_virt(self, phys: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def phys_to_pfn(self, phys: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def pfn_to_phys(self, pfn: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def pfn_to_page(self, pfn: int) -> int:
        raise NotImplementedError()

    @abstractmethod
    def page_to_pfn(self, page: int) -> int:
        raise NotImplementedError()

    def _paginginfo(self) -> pwndbg.aglib.kernel.paging.ArchPagingInfo:
        result = arch_paginginfo()
        if not result:
            raise NotImplementedError()
        return result

    @property
    @pwndbg.lib.cache.cache_until("start")
    def STRUCT_PAGE_SIZE(self):
        return self._paginginfo().STRUCT_PAGE_SIZE

    @property
    @pwndbg.lib.cache.cache_until("start")
    def STRUCT_PAGE_SHIFT(self):
        return self._paginginfo().STRUCT_PAGE_SHIFT

    @property
    def page_offset(self) -> int:
        return self._paginginfo().physmap

    @property
    def phys_offset(self) -> int:
        return self._paginginfo().phys_offset

    @property
    def page_shift(self) -> int:
        return self._paginginfo().page_shift

    @property
    def vmemmap(self) -> int:
        return self._paginginfo().vmemmap

    @property
    def kbase(self) -> int | None:
        return self._paginginfo().kbase

    @property
    def vmalloc(self) -> int:
        return self._paginginfo().vmalloc

    @property
    def page_size(self) -> int:
        return self._paginginfo().page_size

    def virt_to_pfn(self, virt: int) -> int:
        return phys_to_pfn(virt_to_phys(virt))

    def pfn_to_virt(self, pfn: int) -> int:
        return phys_to_virt(pfn_to_phys(pfn))

    def phys_to_page(self, phys: int) -> int:
        return pfn_to_page(phys_to_pfn(phys))

    def page_to_phys(self, page: int) -> int:
        return pfn_to_phys(page_to_pfn(page))

    def virt_to_page(self, virt: int) -> int:
        return pfn_to_page(virt_to_pfn(virt))

    def page_to_virt(self, page: int) -> int:
        return pfn_to_virt(page_to_pfn(page))


class x86Ops(ArchOps):
    def phys_to_virt(self, phys: int) -> int:
        return pwndbg.aglib.arch.unsigned(phys + self.page_offset)

    def phys_to_pfn(self, phys: int) -> int:
        return phys >> self.page_shift

    def pfn_to_phys(self, pfn: int) -> int:
        return pfn << self.page_shift

    @staticmethod
    def paging_enabled() -> bool:
        return int(pwndbg.aglib.regs.read_reg("cr0")) & BIT(31) != 0


class i386Ops(x86Ops):
    def virt_to_phys(self, virt: int) -> int:
        return (virt - self.page_offset) % (1 << 32)

    def per_cpu(
        self, addr: int | pwndbg.dbg_mod.Value, cpu: int | None = None
    ) -> pwndbg.dbg_mod.Value:
        raise NotImplementedError()

    def pfn_to_page(self, pfn: int) -> int:
        raise NotImplementedError()

    def page_to_pfn(self, page: int) -> int:
        raise NotImplementedError()


class x86_64Ops(x86Ops):
    def __init__(self) -> None:
        self.phys_base = 0x1000000

    @requires_debug_symbols("__per_cpu_offset", "nr_iowait_cpu", checkall=False)
    def per_cpu(
        self, addr: int | pwndbg.dbg_mod.Value, cpu: int | None = None
    ) -> pwndbg.dbg_mod.Value:
        if cpu is None:
            cpu = current_cpu()

        per_cpu_offset = int(pwndbg.aglib.kernel.per_cpu_offset())

        offset = pwndbg.aglib.memory.read_pointer_width(per_cpu_offset + (cpu * 8))
        per_cpu_addr = (int(addr) + offset) % 2**64
        if isinstance(addr, pwndbg.dbg_mod.Value):
            return pwndbg.dbg.selected_inferior().create_value(per_cpu_addr, addr.type)
        return pwndbg.dbg.selected_inferior().create_value(per_cpu_addr)

    def virt_to_phys(self, virt: int) -> int:
        _virt: int | None = virt
        if not (pwndbg.aglib.memory.is_kernel(virt) and virt < self.vmalloc):
            # if not within physmap range, first find the physmap address
            _virt = pagewalk(virt).virt
        if _virt is None:
            _virt = virt
        return _virt - self.page_offset

    def pfn_to_page(self, pfn: int) -> int:
        # assumption: SPARSEMEM_VMEMMAP memory model used
        # FLATMEM or SPARSEMEM not (yet) implemented
        return (pfn << self.STRUCT_PAGE_SHIFT) + self.vmemmap

    def page_to_pfn(self, page: int) -> int:
        # assumption: SPARSEMEM_VMEMMAP memory model used
        # FLATMEM or SPARSEMEM not (yet) implemented
        return (page - self.vmemmap) >> self.STRUCT_PAGE_SHIFT


class Aarch64Ops(ArchOps):
    @requires_debug_symbols("__per_cpu_offset", "nr_iowait_cpu", checkall=False)
    def per_cpu(
        self, addr: int | pwndbg.dbg_mod.Value, cpu: int | None = None
    ) -> pwndbg.dbg_mod.Value:
        if cpu is None:
            cpu = current_cpu()

        per_cpu_offset = int(pwndbg.aglib.kernel.per_cpu_offset())

        offset = pwndbg.aglib.memory.u(per_cpu_offset + (cpu * 8))
        per_cpu_addr = (int(addr) + offset) % 2**64
        if isinstance(addr, pwndbg.dbg_mod.Value):
            return pwndbg.dbg.selected_inferior().create_value(per_cpu_addr, addr.type)
        return pwndbg.dbg.selected_inferior().create_value(per_cpu_addr)

    def virt_to_phys(self, virt: int) -> int:
        _virt: int | None = virt
        if not (pwndbg.aglib.memory.is_kernel(virt) and virt < self.vmalloc):
            # if not within physmap range, first find the physmap address
            _virt = pagewalk(virt).virt
        if _virt is None:
            _virt = virt
        return _virt - self.page_offset + self.phys_offset

    def phys_to_virt(self, phys: int) -> int:
        # https://elixir.bootlin.com/linux/v6.16.4/source/arch/arm64/include/asm/memory.h#L356
        return phys - self.phys_offset + self.page_offset

    def phys_to_pfn(self, phys: int) -> int:
        return phys >> self.page_shift

    def pfn_to_phys(self, pfn: int) -> int:
        return pfn << self.page_shift

    def pfn_to_page(self, pfn: int) -> int:
        # assumption: SPARSEMEM_VMEMMAP memory model used
        # FLATMEM or SPARSEMEM not (yet) implemented
        return (pfn << self.STRUCT_PAGE_SHIFT) + self.vmemmap

    def page_to_pfn(self, page: int) -> int:
        # assumption: SPARSEMEM_VMEMMAP memory model used
        # FLATMEM or SPARSEMEM not (yet) implemented
        return (page - self.vmemmap) >> self.STRUCT_PAGE_SHIFT

    @staticmethod
    def paging_enabled() -> bool:
        return int(pwndbg.aglib.regs.read_reg("SCTLR")) & BIT(0) != 0


@pwndbg.lib.cache.cache_until("start")
def arch_paginginfo() -> pwndbg.aglib.kernel.paging.ArchPagingInfo | None:
    if pwndbg.aglib.arch.name == "aarch64":
        return pwndbg.aglib.kernel.paging.Aarch64PagingInfo()
    if pwndbg.aglib.arch.name == "x86-64":
        return pwndbg.aglib.kernel.paging.x86_64PagingInfo()
    return None


@pwndbg.lib.cache.cache_until("start")
def arch_ops() -> ArchOps | None:
    if pwndbg.aglib.arch.name == "aarch64":
        return Aarch64Ops()
    if pwndbg.aglib.arch.name == "x86-64":
        return x86_64Ops()
    if pwndbg.aglib.arch.name == "i386":
        return i386Ops()
    return None


@pwndbg.lib.cache.cache_until("start")
def arch_symbols() -> pwndbg.aglib.kernel.symbol.ArchSymbols | None:
    if pwndbg.aglib.arch.name == "aarch64":
        return pwndbg.aglib.kernel.symbol.Aarch64Symbols()
    if pwndbg.aglib.arch.name == "x86-64":
        return pwndbg.aglib.kernel.symbol.x86_64Symbols()
    return None


def page_size() -> int:
    ops = arch_ops()
    if ops:
        return ops.page_size
    raise NotImplementedError()


def per_cpu(addr: int | pwndbg.dbg_mod.Value, cpu: int | None = None) -> pwndbg.dbg_mod.Value:
    ops = arch_ops()
    if ops:
        return ops.per_cpu(addr, cpu)
    raise NotImplementedError()


def virt_to_phys(virt: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.virt_to_phys(virt)
    raise NotImplementedError()


def phys_to_virt(phys: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.phys_to_virt(phys)
    raise NotImplementedError()


def phys_to_pfn(phys: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.phys_to_pfn(phys)
    raise NotImplementedError()


def pfn_to_phys(pfn: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.pfn_to_phys(pfn)
    raise NotImplementedError()


def pfn_to_page(pfn: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.pfn_to_page(pfn)
    raise NotImplementedError()


def page_to_pfn(page: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.page_to_pfn(page)
    raise NotImplementedError()


def phys_to_page(phys: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.phys_to_page(phys)
    raise NotImplementedError()


def page_to_phys(page: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.page_to_phys(page)
    raise NotImplementedError()


def virt_to_page(virt: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.virt_to_page(virt)
    raise NotImplementedError()


def page_to_virt(page: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.page_to_virt(page)
    raise NotImplementedError()


def pfn_to_virt(pfn: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.pfn_to_virt(pfn)
    raise NotImplementedError()


def virt_to_pfn(virt: int) -> int:
    ops = arch_ops()
    if ops:
        return ops.virt_to_pfn(virt)
    raise NotImplementedError()


@pwndbg.lib.cache.cache_until("stop")
def kbase() -> int | None:
    ops = arch_ops()
    if ops:
        return ops.kbase
    raise NotImplementedError()


@pwndbg.lib.cache.cache_until("stop")
def page_shift() -> int:
    ops = arch_ops()
    if ops:
        return ops.page_shift
    raise NotImplementedError()


@pwndbg.lib.cache.cache_until("stop")
def pagewalk(
    addr, entry: int | None = None, virt: bool = True
) -> pwndbg.aglib.kernel.paging.PagewalkResult:
    """
    assumes entry is a valid physaddr (+ flags)
    the strategy is to walk any virtual pgd first
    """
    pi = arch_paginginfo()
    if pi:
        return pi.pagewalk(addr, entry, virt)
    raise NotImplementedError()


@pwndbg.lib.cache.cache_until("stop")
def pagescan(entry=None) -> tuple[pwndbg.lib.memory.Page, ...]:
    pi = arch_paginginfo()
    if pi:
        return tuple(pi.pagescan(entry))
    raise NotImplementedError()


def bitflags(level: pwndbg.aglib.kernel.paging.PageTableLevel) -> BitFlags:
    pi = arch_paginginfo()
    if pi:
        return pi.bitflags(level)
    raise NotImplementedError()


def slab_to_virt(slab: int) -> int:
    pi = arch_paginginfo()
    if pi:
        return pi.slab_to_virt(slab)
    raise NotImplementedError()


def virt_to_slab(slab: int) -> int:
    pi = arch_paginginfo()
    if pi:
        return pi.virt_to_slab(slab)
    raise NotImplementedError()


def slab_virtual() -> int:
    pi = arch_paginginfo()
    if pi:
        return pi.slab_virtual
    raise NotImplementedError()


def paging_enabled() -> bool:
    arch_name = pwndbg.aglib.arch.name
    if arch_name == "i386":
        return i386Ops.paging_enabled()
    if arch_name == "x86-64":
        return x86_64Ops.paging_enabled()
    if arch_name == "aarch64":
        return Aarch64Ops.paging_enabled()
    if arch_name == "rv64":
        # https://starfivetech.com/uploads/u74_core_complex_manual_21G1.pdf
        # page 41, satp.MODE, bits: 60,61,62,63
        # "When satp.MODE=0x0, supervisor virtual addresses are equal to supervisor physical addresses"
        return (
            int(pwndbg.aglib.regs.read_reg("satp")) & (BIT(60) | BIT(61) | BIT(62) | BIT(63)) != 0
        )
    raise NotImplementedError()


@requires_debug_symbols("node_states", default=1)
def num_numa_nodes() -> int:
    """Returns the number of NUMA nodes that are online on the system"""
    kc = kconfig()

    if "CONFIG_NUMA" not in kc:
        return 1

    if "CONFIG_NODES_SHIFT" not in kc:
        node_states = pwndbg.aglib.symbol.lookup_symbol("node_states")
        if node_states is None or not has_debug_info():
            return 1
        node_states = node_states.dereference()

        # 1 means aglib.typeinfo.enum_member("enum node_states", "N_ONLINE")
        node_mask = node_states[1]["bits"][0]
        return bin(int(node_mask)).count("1")

    max_nodes = 1 << int(kc["CONFIG_NODES_SHIFT"])
    if max_nodes == 1:
        return 1

    val = pwndbg.aglib.kernel.symbol.try_usymbol("nr_online_nodes", 32)
    if val is None:
        return 1

    return val


def node_data() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.node_data()
    return None


def slab_caches() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.slab_caches()
    return None


def per_cpu_offset() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.per_cpu_offset()
    return None


def modules() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.modules()
    return None


def db_list() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.db_list()
    return None


def prog_idr() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.prog_idr()
    return None


def map_idr() -> pwndbg.dbg_mod.Value:
    if (syms := arch_symbols()) is not None:
        return syms.map_idr()
    return None


def current_task(cpu: int | None = None) -> int:
    if (syms := arch_symbols()) is not None:
        return syms.current_task(cpu)
    return None



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

import pwndbg
import pwndbg.aglib.heap.heap
import pwndbg.aglib.proc
import pwndbg.dbg_mod
import pwndbg.lib.config
import pwndbg.libc
from pwndbg.color import message
from pwndbg.dbg_mod import EventType
from pwndbg.lib.config import Scope

current: pwndbg.aglib.heap.heap.MemoryAllocator | None = None


def add_heap_param(
    name: str,
    default: Any,
    set_show_doc: str,
    *,
    help_docstring: str = "",
    param_class: int | None = None,
    enum_sequence: Sequence[str] | None = None,
):
    return pwndbg.config.add_param(
        name,
        default,
        set_show_doc,
        help_docstring=help_docstring,
        param_class=param_class,
        enum_sequence=enum_sequence,
        scope=Scope.heap,
    )


main_arena = add_heap_param("main-arena", "0", "the address of main_arena")

thread_arena = add_heap_param("thread-arena", "0", "the address pointed by thread_arena")

mp_ = add_heap_param("mp", "0", "the address of mp_")

tcache = add_heap_param("tcache", "0", "the address pointed by tcache")

global_max_fast = add_heap_param("global-max-fast", "0", "the address of global_max_fast")

symbol_list = [main_arena, thread_arena, mp_, tcache, global_max_fast]

heap_chain_limit = add_heap_param(
    "heap-dereference-limit",
    8,
    "number of chunks to dereference in each bin",
    param_class=pwndbg.lib.config.PARAM_UINTEGER,
)

heap_corruption_check_limit = add_heap_param(
    "heap-corruption-check-limit",
    64,
    "amount of chunks to traverse for the bin corruption check",
    param_class=pwndbg.lib.config.PARAM_UINTEGER,
    help_docstring="""
The bins are traversed both forwards and backwards.
""",
)

if pwndbg.dbg.name() == pwndbg.dbg_mod.DebuggerType.GDB:
    extra_hint_for_gdb = """
In addition, even you have the debug symbols of libc, you might still see the
following warning when debugging a multi-threaded program:
```
warning: Unable to find libthread_db matching inferior's thread library, thread
debugging will not be available.
```

You'll need to ensure that the correct `libthread_db.so` is loaded. To do this,
set the search path using:
```
set libthread-db-search-path <path having correct libthread_db.so>
```
Then, restart your program to enable proper thread debugging.
"""
else:
    extra_hint_for_gdb = ""

resolve_heap_via_heuristic = add_heap_param(
    "resolve-heap-via-heuristic",
    "auto",
    "the strategy to resolve heap via heuristic",
    help_docstring="""\
Values explained:

+ `auto` - Pwndbg will try to use heuristics if debug symbols are missing
+ `force` - Pwndbg will always try to use heuristics, even if debug symbols are available
+ `never` - Pwndbg will never use heuristics to resolve the heap

If the output of the heap related command produces errors with heuristics, you
can try manually setting the libc symbol addresses.
For this, see the `heap_config` command output and set the `main_arena`, `mp_`,
`global_max_fast`, `tcache` and `thread_arena` addresses.

Note: Pwndbg will generate more reliable results with proper debug symbols.
Therefore, when debug symbols are missing, you should try to install them first
if you haven't already.

They can probably be installed via the package manager of your choice.
See also: https://sourceware.org/gdb/onlinedocs/gdb/Separate-Debug-Files.html .

E.g. on Ubuntu/Debian you might need to do the following steps (for 64-bit and
32-bit binaries):
```bash
sudo apt-get install libc6-dbg
sudo dpkg --add-architecture i386
sudo apt-get install libc-dbg:i386
```
If you used setup.sh on Arch based distro you'll need to do a power cycle or set
environment variable manually like this:
```bash
export DEBUGINFOD_URLS=https://debuginfod.archlinux.org
```
"""
    + extra_hint_for_gdb,
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["auto", "force", "never"],
)
del extra_hint_for_gdb


@pwndbg.dbg.event_handler(EventType.START)
def update() -> None:
    resolve_heap(is_first_run=True)


@pwndbg.dbg.event_handler(EventType.EXIT)
def reset() -> None:
    global current
    # Re-initialize the heap
    if current:
        current = type(current)()
    for symbol in symbol_list:
        symbol.value = "0"


@pwndbg.config.trigger(resolve_heap_via_heuristic)
def resolve_heap(is_first_run: bool = False) -> None:
    global pwndbg
    import pwndbg.aglib.heap.ptmalloc

    global current
    if resolve_heap_via_heuristic == "force":
        current = pwndbg.aglib.heap.ptmalloc.HeuristicHeap()
        if not is_first_run and pwndbg.aglib.proc.alive() and pwndbg.libc.has_debug_info():
            print(
                message.warn(
                    "You are going to resolve the heap via heuristic even though you have libc debug symbols."
                    " This is not recommended!"
                )
            )
    else:
        current = pwndbg.aglib.heap.ptmalloc.DebugSymsHeap()



from __future__ import annotations

from pathlib import Path


def clean_path(path: str) -> str:
    # Why try to avoid: Path("[heap]").resolve() == PosixPath('/home/user/<cwd>/[heap]')
    # FIXME: This is quite flaky and should be standardized in the codebase, see #3641 .
    if not (path.startswith("target:") or path.startswith("[")):
        return str(Path(path).resolve())
    return str(Path(path))



from __future__ import annotations

import argparse
import shutil
import subprocess

import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Parse a struct sock_fprog from memory and dump its filter"
)
parser.add_argument(
    "addr",
    type=int,
    help="Address of sock_fprog structure in target process memory (e.g. 0xdeadbeef)",
)


@pwndbg.commands.Command(parser, command_name="parse-seccomp", category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
def parse_seccomp(addr: int) -> None:
    """Parse a struct sock_fprog at a given address and pass filter to external tool."""

    # addr = int(addr) & pwndbg.aglib.arch.ptrmask
    filter_len = pwndbg.aglib.memory.u16(addr)
    filter_addr = pwndbg.aglib.memory.u(addr + pwndbg.aglib.typeinfo.ptrsize)

    print(message.success(f"sock_fprog @ {addr:#x}"))
    print(f"  len          = {filter_len}")
    print(f"  filter_addr  = {filter_addr:#x}")

    filter_size = filter_len * 8
    filter_bytes = pwndbg.aglib.memory.read(filter_addr, filter_size, partial=False)

    if shutil.which("ceccomp"):
        proc = subprocess.run(
            ["ceccomp", "disasm", "--color", "always"], input=filter_bytes, capture_output=True
        )
        print(proc.stdout.decode())
    elif shutil.which("seccomp-tools"):
        proc = subprocess.run(
            ["seccomp-tools", "disasm", "-"], input=filter_bytes, capture_output=True
        )
        print(proc.stdout.decode())
    else:
        print("install ceccomp or seccomp-tools to parse seccomp")



"""
Command to print the virtual memory map a la /proc/self/maps.
"""

from __future__ import annotations

import argparse
import os.path

from elftools.elf.constants import SH_FLAGS
from elftools.elf.elffile import ELFFile

import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.file
import pwndbg.aglib.vmmap
import pwndbg.aglib.vmmap_custom
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.color import cyan
from pwndbg.color import green
from pwndbg.color import red
from pwndbg.commands import CommandCategory
from pwndbg.lib.memory import Page

integer_types = (int, pwndbg.dbg_mod.Value)


def pages_filter(gdbval_or_str):
    # returns a module filter
    if isinstance(gdbval_or_str, str):
        module_name = gdbval_or_str
        return lambda page: module_name in page.objfile

    # returns an address filter
    if isinstance(gdbval_or_str, integer_types):
        addr = gdbval_or_str
        return lambda page: addr in page

    raise argparse.ArgumentTypeError("Unknown vmmap argument type.")


def print_vmmap_table_header(prefix: str = "") -> None:
    """
    Prints the table header for the vmmap command.
    """
    prefer_relpaths = "on" if pwndbg.config.vmmap_prefer_relpaths else "off"
    width = 2 + 2 * pwndbg.aglib.arch.ptrsize
    print(
        f"{prefix}{'Start':>{width}} {'End':>{width}} {'Perm'} {'Size':>8} {'Offset':>7} "
        f"{'File'} (set vmmap-prefer-relpaths {prefer_relpaths})"
    )


def print_vmmap_gaps_table_header() -> None:
    """
    Prints the table header for the vmmap --gaps command.
    """
    header = (
        f"{'Start':>{2 + 2 * pwndbg.aglib.arch.ptrsize}} "
        f"{'End':>{2 + 2 * pwndbg.aglib.arch.ptrsize}} "
        f"{'Perm':>4} "
        f"{'Size':>8} "
        f"{'Note':>9} "
        f"{'Accumulated Size':>{2 + 2 * pwndbg.aglib.arch.ptrsize}}"
    )
    print(header)


def calculate_total_memory(pages: tuple[Page, ...]) -> None:
    total = 0
    for page in pages:
        total += page.memsz
    if total > 1024 * 1024:
        print(f"Total memory mapped: {total:#x} ({total // 1024 // 1024} MB)")
    else:
        print(f"Total memory mapped: {total:#x} ({total // 1024} KB)")


def gap_text(page: Page) -> str:
    # Strip out offset and objfile from stringified page
    display_text = " ".join(str(page).split(" ")[:-2])
    return display_text.rstrip()


def print_map(page: Page) -> None:
    print(green(gap_text(page)))


def print_adjacent_map(map_start: Page, map_end: Page) -> None:
    print(
        green(
            f"{gap_text(map_end)} {'ADJACENT':>9} {hex(map_end.end - map_start.start):>{2 + 2 * pwndbg.aglib.arch.ptrsize}}"
        )
    )


def print_guard(page: Page) -> None:
    print(cyan(f"{gap_text(page)} {'GUARD':>9} "))


def print_gap(current: Page, last_map: Page):
    print(
        red(
            " - " * int(51 / 3)
            + f" {'GAP':>9} {hex(current.start - last_map.end):>{2 + 2 * pwndbg.aglib.arch.ptrsize}}"
        )
    )


def print_vmmap_gaps(pages: tuple[Page, ...]) -> None:
    """
    Indicates the size of adjacent memory regions and unmapped gaps between them in process memory
    """
    print(f"LEGEND: {green('MAPPED')} | {cyan('GUARD')} | {red('GAP')}")
    print_vmmap_gaps_table_header()

    last_map = None  # The last mapped region we looked at
    last_start = None  # The last starting region of a series of mapped regions

    for page in pages:
        if last_map:
            # If there was a gap print it, and also print the last adjacent map set length
            if last_map.end != page.start:
                if last_start and last_start != last_map:
                    print_adjacent_map(last_start, last_map)
                print_gap(page, last_map)

            # If this is a guard page, print the last map and the guard page
            elif page.is_guard:
                if last_start and last_start != last_map:
                    print_adjacent_map(last_start, last_map)
                print_guard(page)
                last_start = None
                last_map = page
                continue

            # If we are tracking an adjacent set, don't print the current one yet
            elif last_start:
                if last_start != last_map:
                    print_map(last_map)
                last_map = page
                continue

        print_map(page)
        last_start = page
        last_map = page
    calculate_total_memory(pages)


parser = argparse.ArgumentParser(
    description="""Print virtual memory map pages.

Unnamed mappings are named as [anon_%#x] where %#x is high part of their start address. This is useful for filtering with `vmmap` or `search` commands.

Known issues with vmmap:
For QEMU user targets, the QEMU's gdbstub does not provide memory maps information to GDB until [0] is finished & merged. We try to deal with it without parsing the QEMU process' /proc/$pid/maps file, but if our approach fails, we simply create a [0, 0xffff...] vmmap which is not great and may result in lack of proper colors or inability to search memory with the `search` command.

For QEMU kernel, we use gdb-pt-dump that parses page tables from the guest by reading /proc/$pid/mem of QEMU process. If this does not work for you, use `set kernel-vmmap-via-page-tables off` to refer to our old method of reading vmmap info from `monitor info mem` command exposed by QEMU. Note that the latter may be slower and will not give full vmmaps permission information.

For coredump debugging, GDB also lacks all vmmap info but we do our best to get it back by using the `info proc mappings` and `maintenance info sections` commands.

As a last resort, we sometimes try to explore the addresses in CPU registers and if they are readable by GDB, we determine their bounds and create an "<explored>" vmmap. However, this method is slow and is not used on each GDB stop.

Memory pages can also be added manually with the use of vmmap-add, vmmap-clear and vmmap-load commands. This may be useful for bare metal debugging.

[0] https://lore.kernel.org/all/20220221030910.3203063-1-dominik.b.czarnota@gmail.com/""",
)
parser.add_argument(
    "gdbval_or_str",
    type=pwndbg.commands.sloppy_gdb_parse,
    nargs="?",
    default=None,
    help="Address or module name filter",
)
parser.add_argument("-w", "--writable", action="store_true", help="Display writable maps only")
parser.add_argument("-x", "--executable", action="store_true", help="Display executable maps only")
parser.add_argument(
    "-s",
    "--expand-shared-cache",
    action="store_true",
    help="Expand all entries in the DYLD Shared Cache (Darwin only)",
)
parser.add_argument(
    "-A", "--lines-after", type=int, help="Number of pages to display after result", default=1
)
parser.add_argument(
    "-B", "--lines-before", type=int, help="Number of pages to display before result", default=1
)
parser.add_argument(
    "-C", "--context", type=int, help="Number of pages to display around the result"
)
parser.add_argument(
    "--gaps",
    action="store_true",
    help="Display unmapped memory gap information in the memory map.",
)


@pwndbg.commands.Command(
    parser, aliases=["lm", "address", "vprot", "libs"], category=CommandCategory.MEMORY
)
@pwndbg.commands.OnlyWhenRunning
def vmmap(
    gdbval_or_str=None,
    writable=False,
    executable=False,
    lines_after=1,
    lines_before=1,
    context=None,
    gaps=False,
    expand_shared_cache=False,
) -> None:
    lookaround_lines_limit = 64

    # Implement a sane limit
    if context is not None:
        lines_after = lines_before = min(lookaround_lines_limit, context)
    else:
        lines_after = min(lookaround_lines_limit, lines_after)
        lines_before = min(lookaround_lines_limit, lines_before)

    # All displayed pages, including lines after and lines before
    vmmap = pwndbg.aglib.vmmap.get_memory_map()
    total_pages = vmmap.ranges()

    # Filtered memory pages, indicated by a backtrace arrow in results
    filtered_pages = []

    # Only filter when -A and -B arguments are valid
    if gdbval_or_str and lines_after >= 0 and lines_before >= 0:
        # Always expand shared cache on detailed output.
        expand_shared_cache = True

        # Find matching page in memory
        filtered_pages = list(filter(pages_filter(gdbval_or_str), total_pages))
        pages_to_display = []

        for matched_page in filtered_pages:
            # Append matched page
            matched_index = total_pages.index(matched_page)

            # Include number of pages preceeding the matched page
            for before_index in range(0, lines_before + 1):
                # Guard index, and only insert the page if it is not displayed yet
                if (
                    matched_index - before_index >= 0
                    and total_pages[matched_index - before_index] not in pages_to_display
                ):
                    pages_to_display.append(total_pages[matched_index - before_index])

            # Include number of pages proceeding the matched page
            for after_index in range(1, lines_after + 1):
                if (
                    matched_index + after_index < len(total_pages) - 1
                    and total_pages[matched_index + after_index] not in pages_to_display
                ):
                    pages_to_display.append(total_pages[matched_index + after_index])

        # Sort results by address
        total_pages = sorted(pages_to_display, key=lambda page: page.vaddr)

    if not total_pages:
        print("There are no mappings for specified address or module.")
        return

    if gaps:
        print_vmmap_gaps(tuple(total_pages))
        return

    # Determine prefix width for alignment when showing filtered results
    prefix_str = str(pwndbg.config.backtrace_prefix)
    empty_prefix = " " * len(prefix_str) if filtered_pages else None
    header_prefix = f"{empty_prefix} " if filtered_pages else ""

    print(mem_color.legend())
    print_vmmap_table_header(header_prefix)

    shared_cache_first = None
    shared_cache_last = None
    shared_cache_collapsed = 0

    def flush_shared_cache_info():
        nonlocal shared_cache_first
        nonlocal shared_cache_last
        if shared_cache_first is not None and shared_cache_last is not None:
            print(
                pwndbg.lib.memory.format_address(
                    shared_cache_first.start,
                    shared_cache_last.end - shared_cache_first.start,
                    "---p",
                    shared_cache_first.offset,
                    pwndbg.aglib.arch.ptrsize,
                    "[DYLD Shared Cache]",
                )
            )

            shared_cache_first = None
            shared_cache_last = None

    for page in total_pages:
        if (executable and not page.execute) or (writable and not page.write):
            continue

        # Omit ranges from the shared cache if requested.
        if page is not None and page.in_darwin_shared_cache and not expand_shared_cache:
            if shared_cache_first is None:
                shared_cache_first = page
            shared_cache_last = page
            shared_cache_collapsed += 1
            continue
        flush_shared_cache_info()

        backtrace_prefix = empty_prefix
        display_text = str(page)

        if page in filtered_pages:
            # If page was one of the original results, add an arrow for clarity
            backtrace_prefix = prefix_str

            # If the page is the only filtered page, insert offset
            if len(filtered_pages) == 1 and isinstance(gdbval_or_str, integer_types):
                display_text = str(page) + " +0x%x" % (int(gdbval_or_str) - page.vaddr)

        print(mem_color.get(page.vaddr, text=display_text, prefix=backtrace_prefix, page=page))

    flush_shared_cache_info()
    if shared_cache_collapsed > 0:
        print(
            f"[Omitted {shared_cache_collapsed} {'entry' if shared_cache_collapsed == 1 else 'entries'} from the DYLD Shared Cache in total, use '-s' to expand]"
        )
        shared_cache_collapsed = 0

    if vmmap.is_qemu():
        print(
            "\n[QEMU <8.1 target detected - vmmap result might not be accurate; see `help vmmap`]"
        )


parser = argparse.ArgumentParser(description="Add virtual memory map page.")
parser.add_argument("start", type=int, help="Starting virtual address")
parser.add_argument("size", type=int, help="Size of the address space, in bytes")
parser.add_argument(
    "flags",
    nargs="?",
    type=str,
    default="",
    help="Flags set by the ELF file (r - read, w - write, x - executable)",
)
parser.add_argument(
    "offset",
    type=int,
    nargs="?",
    default=0,
    help="Offset into the original ELF file that the data is loaded from",
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def vmmap_add(start: int, size: int, flags: str, offset: int) -> None:
    page_flags = {
        "r": pwndbg.aglib.elf.PF_R,
        "w": pwndbg.aglib.elf.PF_W,
        "x": pwndbg.aglib.elf.PF_X,
    }
    perm = 0
    for flag in flags:
        flag_val = page_flags.get(flag, None)
        if flag_val is None:
            print('Invalid page flag "%s"', flag)
            return
        perm |= flag_val

    page = pwndbg.lib.memory.Page(start, size, perm, offset, pwndbg.aglib.arch.ptrsize)
    pwndbg.aglib.vmmap_custom.add_custom_page(page)

    print(f"{page!r} added")


parser = argparse.ArgumentParser(description="Explore a page, trying to guess permissions.")
parser.add_argument(
    "address", type=pwndbg.commands.sloppy_gdb_parse, help="Address of the page to explore"
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def vmmap_explore(address: int) -> None:
    if not isinstance(address, int):
        print("Address is not a valid integer.")
        return

    old_value = pwndbg.config.auto_explore_pages.value
    pwndbg.config.auto_explore_pages.value = "yes"
    try:
        pwndbg.lib.cache.clear_function_cache(pwndbg.aglib.vmmap.find)
        page = pwndbg.aglib.vmmap.find(address)
    finally:
        pwndbg.config.auto_explore_pages.value = old_value

    if page is None:
        print("Exploration failed. Maybe the address isn't readable?")
        return

    print_vmmap_table_header()
    print(page)


@pwndbg.commands.Command(
    "Clear the vmmap cache.", category=CommandCategory.MEMORY
)  # TODO is this accurate?
@pwndbg.commands.OnlyWhenRunning
def vmmap_clear() -> None:
    pwndbg.aglib.vmmap_custom.clear_custom_page()


parser = argparse.ArgumentParser(description="Load virtual memory map pages from ELF file.")
parser.add_argument(
    "filename",
    nargs="?",
    type=str,
    help="ELF filename, by default uses current loaded filename.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
def vmmap_load(filename) -> None:
    if filename is None:
        filename = pwndbg.aglib.file.get_proc_exe_file()

    print(f'Load "{filename}" ...')
    file_basename = os.path.basename(filename)

    # TODO: Add an argument to let use to choose loading the page information from sections or segments

    # Use section information to recover the segment information.
    # The entry point of bare metal environment is often at the first segment.
    # For example, assume the entry point is at 0x8000.
    # In most of case, link will create a segment and starts from 0x0.
    # This cause all values less than 0x8000 be considered as a valid pointer.
    pages = []
    with open(filename, "rb") as f:
        elffile = ELFFile(f)

        ptrsize: int = pwndbg.aglib.arch.ptrsize
        for section in elffile.iter_sections():
            vaddr = section["sh_addr"]
            memsz = section["sh_size"]
            sh_flags = section["sh_flags"]
            offset = section["sh_offset"]

            # Don't add the sections that aren't mapped into memory
            if not sh_flags & SH_FLAGS.SHF_ALLOC:
                continue

            # Guess the segment flags from section flags
            flags = pwndbg.aglib.elf.PF_R
            if sh_flags & SH_FLAGS.SHF_WRITE:
                flags |= pwndbg.aglib.elf.PF_W
            if sh_flags & SH_FLAGS.SHF_EXECINSTR:
                flags |= pwndbg.aglib.elf.PF_X

            page = pwndbg.lib.memory.Page(
                vaddr, memsz, flags, offset, ptrsize, f"[{section.name}]: {file_basename}"
            )
            pages.append(page)

    for page in pages:
        pwndbg.aglib.vmmap_custom.add_custom_page(page)
        print(f"{page!r} added")



from __future__ import annotations

import importlib.abc
import os
import sys
import traceback

import gdb


def fix_exit():
    major_ver = int(gdb.VERSION.split(".")[0])
    if major_ver <= 15:
        # On certain verions of gdb (used on ubuntu 24.04) using sys.exit() can cause
        # a segfault. See:
        # https://github.com/pwndbg/pwndbg/pull/2900#issuecomment-2825456636
        # https://sourceware.org/bugzilla/show_bug.cgi?id=31946
        def _patched_exit(exit_code):
            # argparse requires a SystemExit exception, otherwise our CLI commands will exit incorrectly on invalid arguments
            stack_list = traceback.extract_stack(limit=2)
            if len(stack_list) == 2:
                p = stack_list[0]
                if p.filename.endswith("/argparse.py"):
                    raise SystemExit()

            sys.stdout.flush()
            sys.stderr.flush()
            os._exit(exit_code)

        sys.exit = _patched_exit


def fix_stdout():
    # Add the original stdout methods back to gdb._GdbOutputFile for pwnlib colors
    sys.stdout.isatty = sys.__stdout__.isatty
    sys.stdout.fileno = sys.__stdout__.fileno


def fix_readline():
    # Fix gdb readline bug: https://github.com/pwndbg/pwndbg/issues/2232#issuecomment-2542564965
    class GdbRemoveReadlineFinder(importlib.abc.MetaPathFinder):
        def find_spec(self, fullname, path=None, target=None):
            if fullname == "readline":
                raise ImportError("readline module disabled under GDB")
            return None

    sys.meta_path.insert(0, GdbRemoveReadlineFinder())


fix_stdout()
fix_readline()
fix_exit()



from __future__ import annotations

import argparse
import subprocess

import pwndbg.aglib
import pwndbg.aglib.elf
import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.radare2
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Launches radare2.")
parser.add_argument("--no-seek", action="store_true", help="Do not seek to current pc")
parser.add_argument(
    "--no-rebase",
    action="store_true",
    help="Do not set the base address for PIE according to the current mapping",
)
parser.add_argument("arguments", nargs="*", type=str, help="Arguments to pass to radare")


@pwndbg.commands.Command(
    parser,
    aliases=["radare2"],
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> r2 -- -S -AA
WARN: Relocs has not been applied. Please use `-e bin.relocs.apply=true` or `-e bin.cache=true` next time
INFO: Analyze all flags starting with sym. and entry0 (aa)
[...]
INFO: Debugger commands disabled in sandbox mode
[0x0001d3d0]> help

Welcome to radare2!
[...]
    """,
)
@pwndbg.commands.OnlyWithFile
def r2(arguments: list[str], no_seek: bool = False, no_rebase: bool = False) -> None:
    filename = pwndbg.aglib.file.get_proc_exe_file()

    # Build up the command line to run
    cmd = ["radare2"]
    flags = ["-e", "io.cache=true"]
    exe = pwndbg.aglib.elf.exe()
    if pwndbg.aglib.proc.alive() and exe is not None:
        addr = pwndbg.aglib.regs.pc
        if pwndbg.aglib.elf.get_elf_info(filename).is_pie:
            if no_rebase:
                addr -= exe.address
            else:
                flags.extend(["-B", hex(exe.address)])
        if not no_seek:
            cmd.extend(["-s", hex(addr)])
    cmd.extend(flags)
    cmd += arguments
    cmd.extend([filename])

    try:
        subprocess.call(cmd)
    except Exception:
        print("Could not run radare2. Please ensure it's installed and in $PATH.")


parser = argparse.ArgumentParser(
    description="Execute stateful radare2 commands through r2pipe.",
)
parser.add_argument("arguments", nargs="+", type=str, help="Arguments to pass to r2pipe")


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> r2pipe pdf @ sym.main
            ; ICOD XREF from entry0 @ 0x1d3e8(r)
┌ 6786: int main (uint32_t argc, char **argv, char **envp);
│ `- args(rdi, rsi, rdx) vars(21:sp[0x40..0x18c])
│           0x0001b920      f30f1efa       endbr64
│           0x0001b924      4157           push r15
│           0x0001b926      4156           push r14
│           0x0001b928      4155           push r13
│           0x0001b92a      4154           push r12
[...]
│           0x0001b966      85c0           test eax, eax
│       ┌─< 0x0001b968      7545           jne 0x1b9af
│       │   0x0001b96a      e8311b0100     call sym.xtrace_init
│       │   0x0001b96f      e80cff0000     call sym.check_dev_tty
│       │   ; CODE XREF from main @ 0x1b988(x)
│      ┌──> 0x0001b974      833dd1680f..   cmp dword [obj.debugging_login_shell], 0 ; [0x11224c:4]=0
│     ┌───< 0x0001b97b      743d           je 0x1b9ba
│     │╎│   0x0001b97d      bf03000000     mov edi, 3
[...]
    """,
)
@pwndbg.commands.OnlyWithFile
def r2pipe(arguments: list[str]) -> None:
    print(pwndbg.aglib.radare2.r2cmd(arguments))
